# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test CF-NetCDF file loading and saving.

"""

# Import iris tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import os
import os.path
import shutil
import stat
from subprocess import check_call
import tempfile
from unittest import mock

from cf_units import as_unit
import netCDF4 as nc
import numpy as np
import numpy.ma as ma

import iris
import iris.analysis.trajectory
import iris.fileformats._pyke_rules.compiled_krb.fc_rules_cf_fc as pyke_rules
import iris.fileformats.netcdf
from iris.fileformats.netcdf import load_cubes as nc_load_cubes
import iris.std_names
import iris.util
from iris.coords import AncillaryVariable, CellMeasure
import iris.coord_systems as icoord_systems
import iris.tests.stock as stock
from iris._lazy_data import is_lazy_data


@tests.skip_data
class TestNetCDFLoad(tests.IrisTest):
    def setUp(self):
        self.tmpdir = None

    def tearDown(self):
        if self.tmpdir is not None:
            shutil.rmtree(self.tmpdir)

    def test_monotonic(self):
        cubes = iris.load(
            tests.get_data_path(
                ("NetCDF", "testing", "test_monotonic_coordinate.nc")
            )
        )
        cubes = sorted(cubes, key=lambda cube: cube.var_name)
        self.assertCML(cubes, ("netcdf", "netcdf_monotonic.cml"))

    def test_load_global_xyt_total(self):
        # Test loading single xyt CF-netCDF file.
        cube = iris.load_cube(
            tests.get_data_path(
                ("NetCDF", "global", "xyt", "SMALL_total_column_co2.nc")
            )
        )
        self.assertCML(cube, ("netcdf", "netcdf_global_xyt_total.cml"))

    def test_load_global_xyt_hires(self):
        # Test loading another single xyt CF-netCDF file.
        cube = iris.load_cube(
            tests.get_data_path(
                ("NetCDF", "global", "xyt", "SMALL_hires_wind_u_for_ipcc4.nc")
            )
        )
        self.assertCML(cube, ("netcdf", "netcdf_global_xyt_hires.cml"))

    def test_missing_time_bounds(self):
        # Check we can cope with a missing bounds variable.
        with self.temp_filename(suffix="nc") as filename:
            # Tweak a copy of the test data file to rename (we can't delete)
            # the time bounds variable.
            src = tests.get_data_path(
                ("NetCDF", "global", "xyt", "SMALL_hires_wind_u_for_ipcc4.nc")
            )
            shutil.copyfile(src, filename)
            dataset = nc.Dataset(filename, mode="a")
            dataset.renameVariable("time_bnds", "foo")
            dataset.close()
            _ = iris.load_cube(filename, "eastward_wind")

    def test_load_global_xyzt_gems(self):
        # Test loading single xyzt CF-netCDF file (multi-cube).
        cubes = iris.load(
            tests.get_data_path(
                ("NetCDF", "global", "xyz_t", "GEMS_CO2_Apr2006.nc")
            )
        )
        cubes = sorted(cubes, key=lambda cube: cube.name())
        self.assertCML(cubes, ("netcdf", "netcdf_global_xyzt_gems.cml"))

        # Check the masked array fill value is propogated through the data
        # manager loading.
        lnsp = cubes[1]
        self.assertTrue(ma.isMaskedArray(lnsp.data))
        self.assertEqual(-32767.0, lnsp.data.fill_value)

    def test_load_global_xyzt_gems_iter(self):
        # Test loading stepped single xyzt CF-netCDF file (multi-cube).
        for i, cube in enumerate(
            sorted(
                iris.load(
                    tests.get_data_path(
                        ("NetCDF", "global", "xyz_t", "GEMS_CO2_Apr2006.nc")
                    )
                ),
                key=lambda cube: cube.name(),
            )
        ):
            self.assertCML(
                cube, ("netcdf", "netcdf_global_xyzt_gems_iter_%d.cml" % i)
            )

    # -------------------------------------------------------------------------
    # It is not considered necessary to have integration tests for
    # loading EVERY coordinate system. A subset are tested below.
    # -------------------------------------------------------------------------

    def test_load_rotated_xy_land(self):
        # Test loading single xy rotated pole CF-netCDF file.
        cube = iris.load_cube(
            tests.get_data_path(
                ("NetCDF", "rotated", "xy", "rotPole_landAreaFraction.nc")
            )
        )
        # Make sure the AuxCoords have lazy data.
        self.assertTrue(is_lazy_data(cube.coord("latitude").core_points()))
        self.assertCML(cube, ("netcdf", "netcdf_rotated_xy_land.cml"))

    def test_load_rotated_xyt_precipitation(self):
        # Test loading single xyt rotated pole CF-netCDF file.
        cube = iris.load_cube(
            tests.get_data_path(
                ("NetCDF", "rotated", "xyt", "small_rotPole_precipitation.nc")
            )
        )
        self.assertCML(
            cube, ("netcdf", "netcdf_rotated_xyt_precipitation.cml")
        )

    def test_load_tmerc_grid_and_clim_bounds(self):
        # Test loading a single CF-netCDF file with a transverse Mercator
        # grid_mapping and a time variable with climatology.
        cube = iris.load_cube(
            tests.get_data_path(
                ("NetCDF", "transverse_mercator", "tmean_1910_1910.nc")
            )
        )
        self.assertCML(cube, ("netcdf", "netcdf_tmerc_and_climatology.cml"))

    def test_load_tmerc_grid_with_projection_origin(self):
        # Test loading a single CF-netCDF file with a transverse Mercator
        # grid_mapping that uses longitude_of_projection_origin and
        # scale_factor_at_projection_origin instead of
        # longitude_of_central_meridian and scale_factor_at_central_meridian.
        cube = iris.load_cube(
            tests.get_data_path(
                (
                    "NetCDF",
                    "transverse_mercator",
                    "projection_origin_attributes.nc",
                )
            )
        )

        expected = icoord_systems.TransverseMercator(
            latitude_of_projection_origin=49.0,
            longitude_of_central_meridian=-2.0,
            false_easting=400000.0,
            false_northing=-100000.0,
            scale_factor_at_central_meridian=0.9996012717,
            ellipsoid=icoord_systems.GeogCS(
                semi_major_axis=6377563.396, semi_minor_axis=6356256.91
            ),
        )
        self.assertEqual(
            cube.coord("projection_x_coordinate").coord_system, expected
        )
        self.assertEqual(
            cube.coord("projection_y_coordinate").coord_system, expected
        )

    def test_load_lcc_grid(self):
        # Test loading a single CF-netCDF file with Lambert conformal conic
        # grid mapping.
        cube = iris.load_cube(
            tests.get_data_path(("NetCDF", "lambert_conformal", "test_lcc.nc"))
        )
        self.assertCML(cube, ("netcdf", "netcdf_lcc.cml"))

    def test_missing_climatology(self):
        # Check we can cope with a missing climatology variable.
        with self.temp_filename(suffix="nc") as filename:
            # Tweak a copy of the test data file to rename (we can't delete)
            # the climatology variable.
            src = tests.get_data_path(
                ("NetCDF", "transverse_mercator", "tmean_1910_1910.nc")
            )
            shutil.copyfile(src, filename)
            dataset = nc.Dataset(filename, mode="a")
            dataset.renameVariable("climatology_bounds", "foo")
            dataset.close()
            _ = iris.load_cube(filename, "Mean temperature")

    def test_load_merc_grid(self):
        # Test loading a single CF-netCDF file with a Mercator grid_mapping
        cube = iris.load_cube(
            tests.get_data_path(
                ("NetCDF", "mercator", "toa_brightness_temperature.nc")
            )
        )
        self.assertCML(cube, ("netcdf", "netcdf_merc.cml"))

    def test_load_stereographic_grid(self):
        # Test loading a single CF-netCDF file with a stereographic
        # grid_mapping.
        cube = iris.load_cube(
            tests.get_data_path(
                ("NetCDF", "stereographic", "toa_brightness_temperature.nc")
            )
        )
        self.assertCML(cube, ("netcdf", "netcdf_stereo.cml"))

    def test_cell_methods(self):
        # Test exercising CF-netCDF cell method parsing.
        cubes = iris.load(
            tests.get_data_path(("NetCDF", "testing", "cell_methods.nc"))
        )

        # TEST_COMPAT mod - new cube merge doesn't sort in the same way - test
        # can pass by manual sorting...
        cubes = iris.cube.CubeList(sorted(cubes, key=lambda cube: cube.name()))

        # TEST_COMPAT mod - different versions of the Python module
        # `netCDF4` give different data arrays: MaskedArray vs ndarray
        # Since we're not interested in the data we can just normalise
        # to MaskedArray (to minimise the change).
        for cube in cubes:
            # Force the fill value to be the default netCDF fill value
            # to ensure it matches the previous behaviour.
            cube.data = ma.masked_equal(cube.data, -2147483647)

        self.assertCML(cubes, ("netcdf", "netcdf_cell_methods.cml"))

    def test_ancillary_variables(self):
        # Note: using a CDL string as a test data reference, rather than a binary file.
        ref_cdl = """
            netcdf cm_attr {
            dimensions:
                axv = 3 ;
            variables:
                int64 qqv(axv) ;
                    qqv:long_name = "qq" ;
                    qqv:units = "1" ;
                    qqv:ancillary_variables = "my_av" ;
                int64 axv(axv) ;
                    axv:units = "1" ;
                    axv:long_name = "x" ;
                double my_av(axv) ;
                    my_av:units = "1" ;
                    my_av:long_name = "refs" ;
                    my_av:custom = "extra-attribute";
            data:
                axv = 1, 2, 3;
                my_av = 11., 12., 13.;
            }
            """
        self.tmpdir = tempfile.mkdtemp()
        cdl_path = os.path.join(self.tmpdir, "tst.cdl")
        nc_path = os.path.join(self.tmpdir, "tst.nc")
        # Write CDL string into a temporary CDL file.
        with open(cdl_path, "w") as f_out:
            f_out.write(ref_cdl)
        # Use ncgen to convert this into an actual (temporary) netCDF file.
        command = "ncgen -o {} {}".format(nc_path, cdl_path)
        check_call(command, shell=True)
        # Load with iris.fileformats.netcdf.load_cubes, and check expected content.
        cubes = list(nc_load_cubes(nc_path))
        self.assertEqual(len(cubes), 1)
        avs = cubes[0].ancillary_variables()
        self.assertEqual(len(avs), 1)
        expected = AncillaryVariable(
            np.ma.array([11.0, 12.0, 13.0]),
            long_name="refs",
            var_name="my_av",
            units="1",
            attributes={"custom": "extra-attribute"},
        )
        self.assertEqual(avs[0], expected)

    def test_status_flags(self):
        # Note: using a CDL string as a test data reference, rather than a binary file.
        ref_cdl = """
            netcdf cm_attr {
            dimensions:
                axv = 3 ;
            variables:
                int64 qqv(axv) ;
                    qqv:long_name = "qq" ;
                    qqv:units = "1" ;
                    qqv:ancillary_variables = "my_av" ;
                int64 axv(axv) ;
                    axv:units = "1" ;
                    axv:long_name = "x" ;
                byte my_av(axv) ;
                    my_av:long_name = "qq status_flag" ;
                    my_av:flag_values = 1b, 2b ;
                    my_av:flag_meanings = "a b" ;
            data:
                axv = 11, 21, 31;
                my_av = 1b, 1b, 2b;
            }
            """
        self.tmpdir = tempfile.mkdtemp()
        cdl_path = os.path.join(self.tmpdir, "tst.cdl")
        nc_path = os.path.join(self.tmpdir, "tst.nc")
        # Write CDL string into a temporary CDL file.
        with open(cdl_path, "w") as f_out:
            f_out.write(ref_cdl)
        # Use ncgen to convert this into an actual (temporary) netCDF file.
        command = "ncgen -o {} {}".format(nc_path, cdl_path)
        check_call(command, shell=True)
        # Load with iris.fileformats.netcdf.load_cubes, and check expected content.
        cubes = list(nc_load_cubes(nc_path))
        self.assertEqual(len(cubes), 1)
        avs = cubes[0].ancillary_variables()
        self.assertEqual(len(avs), 1)
        expected = AncillaryVariable(
            np.ma.array([1, 1, 2], dtype=np.int8),
            long_name="qq status_flag",
            var_name="my_av",
            units="no_unit",
            attributes={
                "flag_values": np.array([1, 2], dtype=np.int8),
                "flag_meanings": "a b",
            },
        )
        self.assertEqual(avs[0], expected)

    def test_cell_measures(self):
        # Note: using a CDL string as a test data reference, rather than a binary file.
        ref_cdl = """
            netcdf cm_attr {
            dimensions:
                axv = 3 ;
                ayv = 2 ;
            variables:
                int64 qqv(ayv, axv) ;
                    qqv:long_name = "qq" ;
                    qqv:units = "1" ;
                    qqv:cell_measures = "area: my_areas" ;
                int64 ayv(ayv) ;
                    ayv:units = "1" ;
                    ayv:long_name = "y" ;
                int64 axv(axv) ;
                    axv:units = "1" ;
                    axv:long_name = "x" ;
                double my_areas(ayv, axv) ;
                    my_areas:units = "m2" ;
                    my_areas:long_name = "standardised cell areas" ;
                    my_areas:custom = "extra-attribute";
            data:
                axv = 11, 12, 13;
                ayv = 21, 22;
                my_areas = 110., 120., 130., 221., 231., 241.;
            }
            """
        self.tmpdir = tempfile.mkdtemp()
        cdl_path = os.path.join(self.tmpdir, "tst.cdl")
        nc_path = os.path.join(self.tmpdir, "tst.nc")
        # Write CDL string into a temporary CDL file.
        with open(cdl_path, "w") as f_out:
            f_out.write(ref_cdl)
        # Use ncgen to convert this into an actual (temporary) netCDF file.
        command = "ncgen -o {} {}".format(nc_path, cdl_path)
        check_call(command, shell=True)
        # Load with iris.fileformats.netcdf.load_cubes, and check expected content.
        cubes = list(nc_load_cubes(nc_path))
        self.assertEqual(len(cubes), 1)
        cms = cubes[0].cell_measures()
        self.assertEqual(len(cms), 1)
        expected = CellMeasure(
            np.ma.array([[110.0, 120.0, 130.0], [221.0, 231.0, 241.0]]),
            measure="area",
            var_name="my_areas",
            long_name="standardised cell areas",
            units="m2",
            attributes={"custom": "extra-attribute"},
        )
        self.assertEqual(cms[0], expected)

    def test_deferred_loading(self):
        # Test exercising CF-netCDF deferred loading and deferred slicing.
        # shape (31, 161, 320)
        cube = iris.load_cube(
            tests.get_data_path(
                ("NetCDF", "global", "xyt", "SMALL_total_column_co2.nc")
            )
        )

        # Consecutive index on same dimension.
        self.assertCML(cube[0], ("netcdf", "netcdf_deferred_index_0.cml"))
        self.assertCML(cube[0][0], ("netcdf", "netcdf_deferred_index_1.cml"))
        self.assertCML(
            cube[0][0][0], ("netcdf", "netcdf_deferred_index_2.cml")
        )

        # Consecutive slice on same dimension.
        self.assertCML(cube[0:20], ("netcdf", "netcdf_deferred_slice_0.cml"))
        self.assertCML(
            cube[0:20][0:10], ("netcdf", "netcdf_deferred_slice_1.cml")
        )
        self.assertCML(
            cube[0:20][0:10][0:5], ("netcdf", "netcdf_deferred_slice_2.cml")
        )

        # Consecutive tuple index on same dimension.
        self.assertCML(
            cube[((0, 8, 4, 2, 14, 12),)],
            ("netcdf", "netcdf_deferred_tuple_0.cml"),
        )
        self.assertCML(
            cube[((0, 8, 4, 2, 14, 12),)][((0, 2, 4, 1),)],
            ("netcdf", "netcdf_deferred_tuple_1.cml"),
        )
        subcube = cube[((0, 8, 4, 2, 14, 12),)][((0, 2, 4, 1),)][
            (1, 3),
        ]
        self.assertCML(subcube, ("netcdf", "netcdf_deferred_tuple_2.cml"))

        # Consecutive mixture on same dimension.
        self.assertCML(
            cube[0:20:2][((9, 5, 8, 0),)][3],
            ("netcdf", "netcdf_deferred_mix_0.cml"),
        )
        self.assertCML(
            cube[((2, 7, 3, 4, 5, 0, 9, 10),)][2:6][3],
            ("netcdf", "netcdf_deferred_mix_0.cml"),
        )
        self.assertCML(
            cube[0][(0, 2), (1, 3)], ("netcdf", "netcdf_deferred_mix_1.cml")
        )

    def test_default_units(self):
        # Note: using a CDL string as a test data reference, rather than a binary file.
        ref_cdl = """
            netcdf cm_attr {
            dimensions:
                axv = 3 ;
                ayv = 2 ;
            variables:
                int64 qqv(ayv, axv) ;
                    qqv:long_name = "qq" ;
                    qqv:ancillary_variables = "my_av" ;
                    qqv:cell_measures = "area: my_areas" ;
                int64 ayv(ayv) ;
                    ayv:long_name = "y" ;
                int64 axv(axv) ;
                    axv:units = "1" ;
                    axv:long_name = "x" ;
                double my_av(axv) ;
                    my_av:long_name = "refs" ;
                double my_areas(ayv, axv) ;
                    my_areas:long_name = "areas" ;
            data:
                axv = 11, 12, 13;
                ayv = 21, 22;
                my_areas = 110., 120., 130., 221., 231., 241.;
            }
            """
        self.tmpdir = tempfile.mkdtemp()
        cdl_path = os.path.join(self.tmpdir, "tst.cdl")
        nc_path = os.path.join(self.tmpdir, "tst.nc")
        # Write CDL string into a temporary CDL file.
        with open(cdl_path, "w") as f_out:
            f_out.write(ref_cdl)
        # Use ncgen to convert this into an actual (temporary) netCDF file.
        command = "ncgen -o {} {}".format(nc_path, cdl_path)
        check_call(command, shell=True)
        # Load with iris.fileformats.netcdf.load_cubes, and check expected content.
        cubes = list(nc_load_cubes(nc_path))
        self.assertEqual(len(cubes), 1)
        self.assertEqual(cubes[0].units, as_unit("unknown"))
        self.assertEqual(cubes[0].coord("y").units, as_unit("unknown"))
        self.assertEqual(cubes[0].coord("x").units, as_unit(1))
        self.assertEqual(
            cubes[0].ancillary_variable("refs").units, as_unit("unknown")
        )
        self.assertEqual(
            cubes[0].cell_measure("areas").units, as_unit("unknown")
        )

    def test_um_stash_source(self):
        """Test that um_stash_source is converted into a STASH code"""
        # Note: using a CDL string as a test data reference, rather than a binary file.
        ref_cdl = """
            netcdf cm_attr {
            dimensions:
                axv = 3 ;
                ayv = 2 ;
            variables:
                int64 qqv(ayv, axv) ;
                    qqv:long_name = "qq" ;
                    qqv:ancillary_variables = "my_av" ;
                    qqv:cell_measures = "area: my_areas" ;
                    qqv:um_stash_source = "m01s02i003" ;
                int64 ayv(ayv) ;
                    ayv:long_name = "y" ;
                int64 axv(axv) ;
                    axv:units = "1" ;
                    axv:long_name = "x" ;
                double my_av(axv) ;
                    my_av:long_name = "refs" ;
                double my_areas(ayv, axv) ;
                    my_areas:long_name = "areas" ;
            data:
                axv = 11, 12, 13;
                ayv = 21, 22;
                my_areas = 110., 120., 130., 221., 231., 241.;
            }
            """
        self.tmpdir = tempfile.mkdtemp()
        cdl_path = os.path.join(self.tmpdir, "tst.cdl")
        nc_path = os.path.join(self.tmpdir, "tst.nc")
        # Write CDL string into a temporary CDL file.
        with open(cdl_path, "w") as f_out:
            f_out.write(ref_cdl)
        # Use ncgen to convert this into an actual (temporary) netCDF file.
        command = "ncgen -o {} {}".format(nc_path, cdl_path)
        check_call(command, shell=True)
        # Load with iris.fileformats.netcdf.load_cubes, and check expected content.
        cubes = list(nc_load_cubes(nc_path))
        self.assertEqual(len(cubes), 1)
        self.assertEqual(
            cubes[0].attributes["STASH"], iris.fileformats.pp.STASH(1, 2, 3)
        )

    def test_ukmo__um_stash_source_priority(self):
        """
        Test that ukmo__um_stash_source is converted into a STASH code with a
        higher priority than um_stash_source.
        """
        # Note: using a CDL string as a test data reference, rather than a binary file.
        ref_cdl = """
            netcdf cm_attr {
            dimensions:
                axv = 3 ;
                ayv = 2 ;
            variables:
                int64 qqv(ayv, axv) ;
                    qqv:long_name = "qq" ;
                    qqv:ancillary_variables = "my_av" ;
                    qqv:cell_measures = "area: my_areas" ;
                    qqv:um_stash_source = "m01s02i003" ;
                    qqv:ukmo__um_stash_source = "m09s08i007" ;
                int64 ayv(ayv) ;
                    ayv:long_name = "y" ;
                int64 axv(axv) ;
                    axv:units = "1" ;
                    axv:long_name = "x" ;
                double my_av(axv) ;
                    my_av:long_name = "refs" ;
                double my_areas(ayv, axv) ;
                    my_areas:long_name = "areas" ;
            data:
                axv = 11, 12, 13;
                ayv = 21, 22;
                my_areas = 110., 120., 130., 221., 231., 241.;
            }
            """
        self.tmpdir = tempfile.mkdtemp()
        cdl_path = os.path.join(self.tmpdir, "tst.cdl")
        nc_path = os.path.join(self.tmpdir, "tst.nc")
        # Write CDL string into a temporary CDL file.
        with open(cdl_path, "w") as f_out:
            f_out.write(ref_cdl)
        # Use ncgen to convert this into an actual (temporary) netCDF file.
        command = "ncgen -o {} {}".format(nc_path, cdl_path)
        check_call(command, shell=True)
        # Load with iris.fileformats.netcdf.load_cubes, and check expected content.
        cubes = list(nc_load_cubes(nc_path))
        self.assertEqual(len(cubes), 1)
        self.assertEqual(
            cubes[0].attributes["STASH"], iris.fileformats.pp.STASH(9, 8, 7)
        )

    def test_bad_um_stash_source(self):
        """Test that um_stash_source not in strict MSI form is kept"""
        # Note: using a CDL string as a test data reference, rather than a binary file.
        ref_cdl = """
            netcdf cm_attr {
            dimensions:
                axv = 3 ;
                ayv = 2 ;
            variables:
                int64 qqv(ayv, axv) ;
                    qqv:long_name = "qq" ;
                    qqv:ancillary_variables = "my_av" ;
                    qqv:cell_measures = "area: my_areas" ;
                    qqv:um_stash_source = "10*m01s02i003" ;
                int64 ayv(ayv) ;
                    ayv:long_name = "y" ;
                int64 axv(axv) ;
                    axv:units = "1" ;
                    axv:long_name = "x" ;
                double my_av(axv) ;
                    my_av:long_name = "refs" ;
                double my_areas(ayv, axv) ;
                    my_areas:long_name = "areas" ;
            data:
                axv = 11, 12, 13;
                ayv = 21, 22;
                my_areas = 110., 120., 130., 221., 231., 241.;
            }
            """
        self.tmpdir = tempfile.mkdtemp()
        cdl_path = os.path.join(self.tmpdir, "tst.cdl")
        nc_path = os.path.join(self.tmpdir, "tst.nc")
        # Write CDL string into a temporary CDL file.
        with open(cdl_path, "w") as f_out:
            f_out.write(ref_cdl)
        # Use ncgen to convert this into an actual (temporary) netCDF file.
        command = "ncgen -o {} {}".format(nc_path, cdl_path)
        check_call(command, shell=True)
        # Load with iris.fileformats.netcdf.load_cubes, and check expected content.
        with self.assertWarns(UserWarning):
            cubes = list(nc_load_cubes(nc_path))
        self.assertEqual(len(cubes), 1)
        self.assertFalse(hasattr(cubes[0].attributes, "STASH"))
        self.assertEqual(
            cubes[0].attributes["um_stash_source"], "10*m01s02i003"
        )

    def test_units(self):
        # Test exercising graceful cube and coordinate units loading.
        cube0, cube1 = sorted(
            iris.load(tests.get_data_path(("NetCDF", "testing", "units.nc"))),
            key=lambda cube: cube.var_name,
        )

        self.assertCML(cube0, ("netcdf", "netcdf_units_0.cml"))
        self.assertCML(cube1, ("netcdf", "netcdf_units_1.cml"))


class TestNetCDFCRS(tests.IrisTest):
    def setUp(self):
        class Var:
            pass

        self.grid = Var()

    def test_lat_lon_major_minor(self):
        major = 63781370
        minor = 63567523
        self.grid.semi_major_axis = major
        self.grid.semi_minor_axis = minor
        crs = pyke_rules.build_coordinate_system(self.grid)
        self.assertEqual(crs, icoord_systems.GeogCS(major, minor))

    def test_lat_lon_earth_radius(self):
        earth_radius = 63700000
        self.grid.earth_radius = earth_radius
        crs = pyke_rules.build_coordinate_system(self.grid)
        self.assertEqual(crs, icoord_systems.GeogCS(earth_radius))


class SaverPermissions(tests.IrisTest):
    def test_noexist_directory(self):
        # Test capture of suitable exception raised on writing to a
        # non-existent directory.
        dir_name = os.path.join(tempfile.gettempdir(), "non_existent_dir")
        fnme = os.path.join(dir_name, "tmp.nc")
        with self.assertRaises(IOError):
            with iris.fileformats.netcdf.Saver(fnme, "NETCDF4"):
                pass

    def test_bad_permissions(self):
        # Skip this test for the root user. This is applicable to
        # running within a Docker container and/or CIaaS hosted testing.
        if os.getuid():
            # Non-exhaustive check that wrong permissions results in a suitable
            # exception being raised.
            dir_name = tempfile.mkdtemp()
            fname = os.path.join(dir_name, "tmp.nc")
            try:
                os.chmod(dir_name, stat.S_IREAD)
                with self.assertRaises(PermissionError):
                    iris.fileformats.netcdf.Saver(fname, "NETCDF4")
                self.assertFalse(os.path.exists(fname))
            finally:
                shutil.rmtree(dir_name)


@tests.skip_data
class TestSave(tests.IrisTest):
    def test_hybrid(self):
        cube = stock.realistic_4d()

        # Write Cube to netCDF file.
        with self.temp_filename(suffix=".nc") as file_out:
            iris.save(cube, file_out, netcdf_format="NETCDF3_CLASSIC")

            # Check the netCDF file against CDL expected output.
            self.assertCDL(
                file_out, ("netcdf", "netcdf_save_realistic_4d.cdl")
            )

    def test_no_hybrid(self):
        cube = stock.realistic_4d()
        cube.remove_aux_factory(cube.aux_factories[0])

        # Write Cube to netCDF file.
        with self.temp_filename(suffix=".nc") as file_out:
            iris.save(cube, file_out, netcdf_format="NETCDF3_CLASSIC")

            # Check the netCDF file against CDL expected output.
            self.assertCDL(
                file_out, ("netcdf", "netcdf_save_realistic_4d_no_hybrid.cdl")
            )

    def test_scalar_cube(self):
        cube = stock.realistic_4d()[0, 0, 0, 0]

        with self.temp_filename(suffix=".nc") as filename:
            iris.save(cube, filename, netcdf_format="NETCDF3_CLASSIC")
            self.assertCDL(
                filename, ("netcdf", "netcdf_save_realistic_0d.cdl")
            )

    def test_no_name_cube(self):
        # Cube with no names.
        cube = iris.cube.Cube(np.arange(20, dtype=np.float64).reshape((4, 5)))
        dim0 = iris.coords.DimCoord(np.arange(4, dtype=np.float64))
        dim1 = iris.coords.DimCoord(np.arange(5, dtype=np.float64), units="m")
        other = iris.coords.AuxCoord("foobar", units="no_unit")
        cube.add_dim_coord(dim0, 0)
        cube.add_dim_coord(dim1, 1)
        cube.add_aux_coord(other)
        with self.temp_filename(suffix=".nc") as filename:
            iris.save(cube, filename, netcdf_format="NETCDF3_CLASSIC")
            self.assertCDL(filename, ("netcdf", "netcdf_save_no_name.cdl"))


class TestNetCDFSave(tests.IrisTest):
    def setUp(self):
        self.cubell = iris.cube.Cube(
            np.arange(4).reshape(2, 2), "air_temperature"
        )
        self.cube = iris.cube.Cube(
            np.zeros([2, 2]),
            standard_name="surface_temperature",
            long_name=None,
            var_name="temp",
            units="K",
        )
        self.cube2 = iris.cube.Cube(
            np.ones([1, 2, 2]),
            standard_name=None,
            long_name="Something Random",
            var_name="temp2",
            units="K",
        )
        self.cube3 = iris.cube.Cube(
            np.ones([2, 2, 2]),
            standard_name=None,
            long_name="Something Random",
            var_name="temp3",
            units="K",
        )
        self.cube4 = iris.cube.Cube(
            np.zeros([10]),
            standard_name="air_temperature",
            long_name=None,
            var_name="temp",
            units="K",
        )
        self.cube5 = iris.cube.Cube(
            np.ones([20]),
            standard_name=None,
            long_name="air_temperature",
            var_name="temp2",
            units="K",
        )
        self.cube6 = iris.cube.Cube(
            np.ones([10]),
            standard_name=None,
            long_name="air_temperature",
            var_name="temp3",
            units="K",
        )

    @tests.skip_data
    def test_netcdf_save_format(self):
        # Read netCDF input file.
        file_in = tests.get_data_path(
            ("NetCDF", "global", "xyt", "SMALL_total_column_co2.nc")
        )
        cube = iris.load_cube(file_in)

        with self.temp_filename(suffix=".nc") as file_out:
            # Test default NETCDF4 file format saving.
            iris.save(cube, file_out)
            ds = nc.Dataset(file_out)
            self.assertEqual(
                ds.file_format, "NETCDF4", "Failed to save as NETCDF4 format"
            )
            ds.close()

            # Test NETCDF4_CLASSIC file format saving.
            iris.save(cube, file_out, netcdf_format="NETCDF4_CLASSIC")
            ds = nc.Dataset(file_out)
            self.assertEqual(
                ds.file_format,
                "NETCDF4_CLASSIC",
                "Failed to save as NETCDF4_CLASSIC format",
            )
            ds.close()

            # Test NETCDF3_CLASSIC file format saving.
            iris.save(cube, file_out, netcdf_format="NETCDF3_CLASSIC")
            ds = nc.Dataset(file_out)
            self.assertEqual(
                ds.file_format,
                "NETCDF3_CLASSIC",
                "Failed to save as NETCDF3_CLASSIC format",
            )
            ds.close()

            # Test NETCDF4_64BIT file format saving.
            iris.save(cube, file_out, netcdf_format="NETCDF3_64BIT")
            ds = nc.Dataset(file_out)
            self.assertTrue(
                ds.file_format in ["NETCDF3_64BIT", "NETCDF3_64BIT_OFFSET"],
                "Failed to save as NETCDF3_64BIT format",
            )
            ds.close()

            # Test invalid file format saving.
            with self.assertRaises(ValueError):
                iris.save(cube, file_out, netcdf_format="WIBBLE")

    @tests.skip_data
    def test_netcdf_save_single(self):
        # Test saving a single CF-netCDF file.
        # Read PP input file.
        file_in = tests.get_data_path(
            (
                "PP",
                "cf_processing",
                "000003000000.03.236.000128.1990.12.01.00.00.b.pp",
            )
        )
        cube = iris.load_cube(file_in)

        # Write Cube to netCDF file.
        with self.temp_filename(suffix=".nc") as file_out:
            iris.save(cube, file_out)

            # Check the netCDF file against CDL expected output.
            self.assertCDL(file_out, ("netcdf", "netcdf_save_single.cdl"))

    # TODO investigate why merge now make time an AuxCoord rather than a
    # DimCoord and why forecast_period is 'preferred'.
    @tests.skip_data
    def test_netcdf_save_multi2multi(self):
        # Test saving multiple CF-netCDF files.
        # Read PP input file.
        file_in = tests.get_data_path(
            ("PP", "cf_processing", "abcza_pa19591997_daily_29.b.pp")
        )
        cubes = iris.load(file_in)

        # Save multiple cubes to multiple files.
        for index, cube in enumerate(cubes):
            # Write Cube to netCDF file.
            with self.temp_filename(suffix=".nc") as file_out:
                iris.save(cube, file_out)

                # Check the netCDF file against CDL expected output.
                self.assertCDL(
                    file_out, ("netcdf", "netcdf_save_multi_%d.cdl" % index)
                )

    @tests.skip_data
    def test_netcdf_save_multi2single(self):
        # Test saving multiple cubes to a single CF-netCDF file.
        # Read PP input file.
        file_in = tests.get_data_path(
            ("PP", "cf_processing", "abcza_pa19591997_daily_29.b.pp")
        )
        cubes = iris.load(file_in)

        # Write Cube to netCDF file.
        with self.temp_filename(suffix=".nc") as file_out:
            # Check that it is the same on loading
            iris.save(cubes, file_out)

            # Check the netCDF file against CDL expected output.
            self.assertCDL(file_out, ("netcdf", "netcdf_save_multiple.cdl"))

    def test_netcdf_multi_nocoord(self):
        # Testing the saving of a cublist with no coords.
        cubes = iris.cube.CubeList([self.cube, self.cube2, self.cube3])
        with self.temp_filename(suffix=".nc") as file_out:
            iris.save(cubes, file_out)

            # Check the netCDF file against CDL expected output.
            self.assertCDL(file_out, ("netcdf", "netcdf_save_nocoord.cdl"))

    def test_netcdf_multi_samevarnme(self):
        # Testing the saving of a cublist with cubes of the same var_name.
        self.cube2.var_name = self.cube.var_name
        cubes = iris.cube.CubeList([self.cube, self.cube2])
        with self.temp_filename(suffix=".nc") as file_out:
            iris.save(cubes, file_out)

            # Check the netCDF file against CDL expected output.
            self.assertCDL(file_out, ("netcdf", "netcdf_save_samevar.cdl"))

    def test_netcdf_multi_with_coords(self):
        # Testing the saving of a cublist with coordinates.
        lat = iris.coords.DimCoord(
            np.arange(2), long_name=None, var_name="lat", units="degree_north"
        )
        lon = iris.coords.DimCoord(
            np.arange(2),
            standard_name="longitude",
            long_name=None,
            var_name="lon",
            units="degree_east",
        )
        rcoord = iris.coords.DimCoord(
            np.arange(1),
            standard_name=None,
            long_name="Rnd Coordinate",
            units=None,
        )
        self.cube.add_dim_coord(lon, 0)
        self.cube.add_dim_coord(lat, 1)
        self.cube2.add_dim_coord(lon, 1)
        self.cube2.add_dim_coord(lat, 2)
        self.cube2.add_dim_coord(rcoord, 0)

        cubes = iris.cube.CubeList([self.cube, self.cube2])
        with self.temp_filename(suffix=".nc") as file_out:
            iris.save(cubes, file_out)

            # Check the netCDF file against CDL expected output.
            self.assertCDL(file_out, ("netcdf", "netcdf_save_wcoord.cdl"))

    def test_netcdf_multi_wtih_samedimcoord(self):
        time1 = iris.coords.DimCoord(
            np.arange(10), standard_name="time", var_name="time", units="1"
        )
        time2 = iris.coords.DimCoord(
            np.arange(20), standard_name="time", var_name="time", units="1"
        )

        self.cube4.add_dim_coord(time1, 0)
        self.cube5.add_dim_coord(time2, 0)
        self.cube6.add_dim_coord(time1, 0)

        cubes = iris.cube.CubeList([self.cube4, self.cube5, self.cube6])
        with self.temp_filename(suffix=".nc") as file_out:
            iris.save(cubes, file_out)

            # Check the netCDF file against CDL expected output.
            self.assertCDL(
                file_out, ("netcdf", "netcdf_save_samedimcoord.cdl")
            )

    def test_netcdf_multi_conflict_name_dup_coord(self):
        # Duplicate coordinates with modified variable names lookup.
        latitude1 = iris.coords.DimCoord(
            np.arange(10), standard_name="latitude", units="1"
        )
        time2 = iris.coords.DimCoord(
            np.arange(2), standard_name="time", units="1"
        )
        latitude2 = iris.coords.DimCoord(
            np.arange(2), standard_name="latitude", units="1"
        )

        self.cube6.add_dim_coord(latitude1, 0)
        self.cube.add_dim_coord(latitude2[:], 1)
        self.cube.add_dim_coord(time2[:], 0)

        cubes = iris.cube.CubeList([self.cube, self.cube6, self.cube6.copy()])
        with self.temp_filename(suffix=".nc") as file_out:
            iris.save(cubes, file_out)

            # Check the netCDF file against CDL expected output.
            self.assertCDL(
                file_out, ("netcdf", "multi_dim_coord_slightly_different.cdl")
            )

    @tests.skip_data
    def test_netcdf_hybrid_height(self):
        # Test saving a CF-netCDF file which contains a hybrid height
        # (i.e. dimensionless vertical) coordinate.
        # Read PP input file.
        names = ["air_potential_temperature", "surface_altitude"]
        file_in = tests.get_data_path(
            ("PP", "COLPEX", "small_colpex_theta_p_alt.pp")
        )
        cube = iris.load_cube(file_in, names[0])

        # Write Cube to netCDF file.
        with self.temp_filename(suffix=".nc") as file_out:
            iris.save(cube, file_out)

            # Check the netCDF file against CDL expected output.
            self.assertCDL(
                file_out, ("netcdf", "netcdf_save_hybrid_height.cdl")
            )

            # Read netCDF file.
            cubes = iris.load(file_out)
            cubes_names = [c.name() for c in cubes]
            self.assertEqual(cubes_names, names)

            # Check the PP read, netCDF write, netCDF read mechanism.
            self.assertCML(
                cubes.extract(names[0])[0],
                ("netcdf", "netcdf_save_load_hybrid_height.cml"),
            )

    @tests.skip_data
    def test_netcdf_save_ndim_auxiliary(self):
        # Test saving CF-netCDF with multi-dimensional auxiliary coordinates.
        # Read netCDF input file.
        file_in = tests.get_data_path(
            ("NetCDF", "rotated", "xyt", "small_rotPole_precipitation.nc")
        )
        cube = iris.load_cube(file_in)

        # Write Cube to nerCDF file.
        with self.temp_filename(suffix=".nc") as file_out:
            iris.save(cube, file_out)

            # Check the netCDF file against CDL expected output.
            self.assertCDL(
                file_out, ("netcdf", "netcdf_save_ndim_auxiliary.cdl")
            )

            # Read the netCDF file.
            cube = iris.load_cube(file_out)

            # Check the netCDF read, write, read mechanism.
            self.assertCML(
                cube, ("netcdf", "netcdf_save_load_ndim_auxiliary.cml")
            )

    def test_netcdf_save_conflicting_aux(self):
        # Test saving CF-netCDF with multi-dimensional auxiliary coordinates,
        # with conflicts.
        self.cube4.add_aux_coord(
            iris.coords.AuxCoord(np.arange(10), "time", units="1"), 0
        )
        self.cube6.add_aux_coord(
            iris.coords.AuxCoord(np.arange(10, 20), "time", units="1"), 0
        )

        cubes = iris.cube.CubeList([self.cube4, self.cube6])
        with self.temp_filename(suffix=".nc") as file_out:
            iris.save(cubes, file_out)

            # Check the netCDF file against CDL expected output.
            self.assertCDL(file_out, ("netcdf", "netcdf_save_conf_aux.cdl"))

    def test_netcdf_save_gridmapping(self):
        # Test saving CF-netCDF from a cubelist with various grid mappings.

        c1 = self.cubell
        c2 = self.cubell.copy()
        c3 = self.cubell.copy()

        coord_system = icoord_systems.GeogCS(6371229)
        coord_system2 = icoord_systems.GeogCS(6371228)
        coord_system3 = icoord_systems.RotatedGeogCS(30, 30)

        c1.add_dim_coord(
            iris.coords.DimCoord(
                np.arange(1, 3),
                "latitude",
                long_name="1",
                units="degrees",
                coord_system=coord_system,
            ),
            1,
        )
        c1.add_dim_coord(
            iris.coords.DimCoord(
                np.arange(1, 3),
                "longitude",
                long_name="1",
                units="degrees",
                coord_system=coord_system,
            ),
            0,
        )

        c2.add_dim_coord(
            iris.coords.DimCoord(
                np.arange(1, 3),
                "latitude",
                long_name="2",
                units="degrees",
                coord_system=coord_system2,
            ),
            1,
        )
        c2.add_dim_coord(
            iris.coords.DimCoord(
                np.arange(1, 3),
                "longitude",
                long_name="2",
                units="degrees",
                coord_system=coord_system2,
            ),
            0,
        )

        c3.add_dim_coord(
            iris.coords.DimCoord(
                np.arange(1, 3),
                "grid_latitude",
                long_name="3",
                units="degrees",
                coord_system=coord_system3,
            ),
            1,
        )
        c3.add_dim_coord(
            iris.coords.DimCoord(
                np.arange(1, 3),
                "grid_longitude",
                long_name="3",
                units="degrees",
                coord_system=coord_system3,
            ),
            0,
        )

        cubes = iris.cube.CubeList([c1, c2, c3])
        with self.temp_filename(suffix=".nc") as file_out:
            iris.save(cubes, file_out)

            # Check the netCDF file against CDL expected output.
            self.assertCDL(
                file_out, ("netcdf", "netcdf_save_gridmapmulti.cdl")
            )

    def test_netcdf_save_conflicting_names(self):
        # Test saving CF-netCDF with a dimension name corresponding to
        # an existing variable name (conflict).
        self.cube4.add_dim_coord(
            iris.coords.DimCoord(np.arange(10), "time", units="1"), 0
        )
        self.cube6.add_aux_coord(
            iris.coords.AuxCoord(1, "time", units="1"), None
        )

        cubes = iris.cube.CubeList([self.cube4, self.cube6])
        with self.temp_filename(suffix=".nc") as file_out:
            iris.save(cubes, file_out)

            # Check the netCDF file against CDL expected output.
            self.assertCDL(file_out, ("netcdf", "netcdf_save_conf_name.cdl"))

    @tests.skip_data
    def test_trajectory(self):
        file_in = tests.get_data_path(("PP", "aPPglob1", "global.pp"))
        cube = iris.load_cube(file_in)

        # extract a trajectory
        xpoint = cube.coord("longitude").points[:10]
        ypoint = cube.coord("latitude").points[:10]
        sample_points = [("latitude", xpoint), ("longitude", ypoint)]
        traj = iris.analysis.trajectory.interpolate(cube, sample_points)

        # save, reload and check
        with self.temp_filename(suffix=".nc") as temp_filename:
            iris.save(traj, temp_filename)
            reloaded = iris.load_cube(temp_filename)
            self.assertCML(
                reloaded, ("netcdf", "save_load_traj.cml"), checksum=False
            )
            self.assertArrayEqual(traj.data, reloaded.data)

    def test_attributes(self):
        # Should be global attributes.
        aglobals = {
            "history": "A long time ago...",
            "title": "Attribute test",
            "foo": "bar",
        }
        for k, v in aglobals.items():
            self.cube.attributes[k] = v
        # Should be overriden.
        aover = {"Conventions": "TEST"}
        for k, v in aover.items():
            self.cube.attributes[k] = v
        # Should be data varible attributes.
        avars = {
            "standard_error_multiplier": 23,
            "flag_masks": "a",
            "flag_meanings": "b",
            "flag_values": "c",
            "missing_value": 1.0e20,
            "STASH": iris.fileformats.pp.STASH(1, 2, 3),
        }
        for k, v in avars.items():
            self.cube.attributes[k] = v
        with self.temp_filename(suffix=".nc") as filename:
            iris.save(self.cube, filename)
            # Load the dataset.
            ds = nc.Dataset(filename, "r")
            exceptions = []
            # Should be global attributes.
            for gkey in aglobals:
                if getattr(ds, gkey) != aglobals.get(gkey):
                    exceptions.append(
                        "{} != {}".format(
                            getattr(ds, gkey), aglobals.get(gkey)
                        )
                    )
            # Should be overriden.
            for okey in aover:
                if getattr(ds, okey) == aover.get(okey):
                    exceptions.append(
                        "{} != {}".format(getattr(ds, okey), avars.get(okey))
                    )
            dv = ds["temp"]
            # Should be data varible attributes;
            # except STASH -> um_stash_source.
            for vkey in avars:
                if vkey != "STASH" and (getattr(dv, vkey) != avars.get(vkey)):
                    exceptions.append(
                        "{} != {}".format(getattr(dv, vkey), avars.get(vkey))
                    )
            if getattr(dv, "um_stash_source") != avars.get("STASH"):
                exc = "{} != {}".format(
                    getattr(dv, "um_stash_source"), avars.get(vkey)
                )
                exceptions.append(exc)
        self.assertEqual(exceptions, [])

    def test_conflicting_attributes(self):
        # Should be data variable attributes.
        self.cube.attributes["foo"] = "bar"
        self.cube2.attributes["foo"] = "orange"
        with self.temp_filename(suffix=".nc") as filename:
            iris.save([self.cube, self.cube2], filename)
            self.assertCDL(filename, ("netcdf", "netcdf_save_confl_attr.cdl"))

    def test_conflicting_global_attributes(self):
        # Should be data variable attributes, but raise a warning.
        attr_name = "history"
        self.cube.attributes[attr_name] = "Team A won."
        self.cube2.attributes[attr_name] = "Team B won."
        expected_msg = (
            "{attr_name!r} is being added as CF data variable "
            "attribute, but {attr_name!r} should only be a CF "
            "global attribute.".format(attr_name=attr_name)
        )
        with self.temp_filename(suffix=".nc") as filename:
            with mock.patch("warnings.warn") as warn:
                iris.save([self.cube, self.cube2], filename)
                warn.assert_called_with(expected_msg)
                self.assertCDL(
                    filename, ("netcdf", "netcdf_save_confl_global_attr.cdl")
                )

    def test_no_global_attributes(self):
        # Should all be data variable attributes.
        # Different keys.
        self.cube.attributes["a"] = "a"
        self.cube2.attributes["b"] = "a"
        self.cube3.attributes["c"] = "a"
        self.cube4.attributes["d"] = "a"
        self.cube5.attributes["e"] = "a"
        self.cube6.attributes["f"] = "a"
        # Different values.
        self.cube.attributes["g"] = "p"
        self.cube2.attributes["g"] = "q"
        self.cube3.attributes["g"] = "r"
        self.cube4.attributes["g"] = "s"
        self.cube5.attributes["g"] = "t"
        self.cube6.attributes["g"] = "u"
        # One different value.
        self.cube.attributes["h"] = "v"
        self.cube2.attributes["h"] = "v"
        self.cube3.attributes["h"] = "v"
        self.cube4.attributes["h"] = "w"
        self.cube5.attributes["h"] = "v"
        self.cube6.attributes["h"] = "v"
        cubes = [
            self.cube,
            self.cube2,
            self.cube3,
            self.cube4,
            self.cube5,
            self.cube6,
        ]
        with self.temp_filename(suffix=".nc") as filename:
            iris.save(cubes, filename)
            self.assertCDL(
                filename, ("netcdf", "netcdf_save_no_global_attr.cdl")
            )


class TestNetCDFSave__ancillaries(tests.IrisTest):
    """Test for saving data with ancillary variables."""

    def test_fulldims(self):
        testcube = stock.realistic_3d()
        ancil = iris.coords.AncillaryVariable(
            np.zeros(testcube.shape),
            long_name="ancil_data",
            units=1,
            attributes={"attr_1": 7, "attr_2": "chat"},
        )
        testcube.add_ancillary_variable(ancil, (0, 1, 2))
        with self.temp_filename(suffix=".nc") as filename:
            iris.save(testcube, filename)
            self.assertCDL(filename)

    def test_partialdims(self):
        # Test saving ancillary data which maps only dims 0 and 2.
        testcube = stock.realistic_3d()
        ancil = iris.coords.AncillaryVariable(
            np.zeros(testcube[:, 0, :].shape),
            long_name="time_lon_values",
            units="m",
        )
        testcube.add_ancillary_variable(ancil, (0, 2))
        with self.temp_filename(suffix=".nc") as filename:
            iris.save(testcube, filename)
            self.assertCDL(filename)

    def test_multiple(self):
        # Test saving with multiple ancillary variables.
        testcube = stock.realistic_3d()
        ancil1_time_lat_lon = iris.coords.AncillaryVariable(
            np.zeros(testcube.shape), long_name="data_values", units=1
        )
        testcube.add_ancillary_variable(ancil1_time_lat_lon, (0, 1, 2))
        ancil2_time = iris.coords.AncillaryVariable(
            np.zeros(testcube[:, 0, 0].shape),
            long_name="time_values",
            units="s",
        )
        testcube.add_ancillary_variable(ancil2_time, 0)
        ancil3_lon = iris.coords.AncillaryVariable(
            np.zeros(testcube[0, 0, :].shape),
            long_name="lon_values",
            units="m",
        )
        testcube.add_ancillary_variable(ancil3_lon, 2)
        with self.temp_filename(suffix=".nc") as filename:
            iris.save(testcube, filename)
            self.assertCDL(filename)

    def test_shared(self):
        # Check that saving cubes with matching ancillaries maps to a shared
        # file variable.
        testcube_1 = stock.realistic_3d()
        ancil = iris.coords.AncillaryVariable(
            np.zeros(testcube_1[0].shape), long_name="latlon_refs", units="s"
        )
        testcube_1.add_ancillary_variable(ancil, (1, 2))

        testcube_2 = testcube_1.copy()
        testcube_2.units = "m"
        testcube_2.rename("alternate_data")
        with self.temp_filename(suffix=".nc") as filename:
            iris.save([testcube_1, testcube_2], filename)
            self.assertCDL(filename)

            # Also check that only one, shared ancillary variable was written.
            ds = nc.Dataset(filename)
            self.assertIn("air_potential_temperature", ds.variables)
            self.assertIn("alternate_data", ds.variables)
            self.assertEqual(
                ds.variables["air_potential_temperature"].ancillary_variables,
                "latlon_refs",
            )
            self.assertEqual(
                ds.variables["alternate_data"].ancillary_variables,
                "latlon_refs",
            )

    def test_aliases(self):
        # Check that saving cubes with *differing* ancillaries of the same name
        # is correctly resolved.
        testcube_1 = stock.realistic_3d()
        testcube_2 = testcube_1.copy()
        testcube_2.units = "m"
        testcube_2.rename("alternate_data")

        ancil1 = iris.coords.AncillaryVariable(
            np.zeros(testcube_1[0].shape), long_name="latlon_refs", units="s"
        )
        testcube_1.add_ancillary_variable(ancil1, (1, 2))

        ancil2 = ancil1.copy()
        ancil2.data[0, 0] += 1.0
        testcube_2.add_ancillary_variable(ancil2, (1, 2))
        with self.temp_filename(suffix=".nc") as filename:
            iris.save([testcube_1, testcube_2], filename)
            self.assertCDL(filename)

    def test_flag(self):
        testcube = stock.realistic_3d()
        flag = iris.coords.AncillaryVariable(
            np.ones(testcube.shape, dtype=np.int8),
            long_name="quality_flag",
            attributes={
                "flag_meanings": "PASS FAIL MISSING",
                "flag_values": np.array([1, 2, 9], dtype=np.int8),
            },
        )
        testcube.add_ancillary_variable(flag, (0, 1, 2))
        with self.temp_filename(suffix=".nc") as filename:
            iris.save(testcube, filename)
            self.assertCDL(filename)


class TestNetCDF3SaveInteger(tests.IrisTest):
    def setUp(self):
        self.cube = iris.cube.Cube(
            np.zeros((2, 2), dtype=np.float64),
            standard_name="surface_temperature",
            long_name=None,
            var_name="temp",
            units="K",
        )

    def test_int64_dimension_coord_netcdf3(self):
        coord = iris.coords.DimCoord(
            np.array([1, 2], dtype=np.int64), long_name="x"
        )
        self.cube.add_dim_coord(coord, 0)
        with self.temp_filename(suffix=".nc") as filename:
            iris.save(self.cube, filename, netcdf_format="NETCDF3_CLASSIC")
            reloaded = iris.load_cube(filename)
            self.assertCML(
                reloaded,
                ("netcdf", "int64_dimension_coord_netcdf3.cml"),
                checksum=False,
            )

    def test_int64_auxiliary_coord_netcdf3(self):
        coord = iris.coords.AuxCoord(
            np.array([1, 2], dtype=np.int64), long_name="x"
        )
        self.cube.add_aux_coord(coord, 0)
        with self.temp_filename(suffix=".nc") as filename:
            iris.save(self.cube, filename, netcdf_format="NETCDF3_CLASSIC")
            reloaded = iris.load_cube(filename)
            self.assertCML(
                reloaded,
                ("netcdf", "int64_auxiliary_coord_netcdf3.cml"),
                checksum=False,
            )

    def test_int64_data_netcdf3(self):
        self.cube.data = self.cube.data.astype(np.int64)
        with self.temp_filename(suffix=".nc") as filename:
            iris.save(self.cube, filename, netcdf_format="NETCDF3_CLASSIC")
            reloaded = iris.load_cube(filename)
            self.assertCML(reloaded, ("netcdf", "int64_data_netcdf3.cml"))

    def test_uint32_dimension_coord_netcdf3(self):
        coord = iris.coords.DimCoord(
            np.array([1, 2], dtype=np.uint32), long_name="x"
        )
        self.cube.add_dim_coord(coord, 0)
        with self.temp_filename(suffix=".nc") as filename:
            iris.save(self.cube, filename, netcdf_format="NETCDF3_CLASSIC")
            reloaded = iris.load_cube(filename)
            self.assertCML(
                reloaded,
                ("netcdf", "uint32_dimension_coord_netcdf3.cml"),
                checksum=False,
            )

    def test_uint32_auxiliary_coord_netcdf3(self):
        coord = iris.coords.AuxCoord(
            np.array([1, 2], dtype=np.uint32), long_name="x"
        )
        self.cube.add_aux_coord(coord, 0)
        with self.temp_filename(suffix=".nc") as filename:
            iris.save(self.cube, filename, netcdf_format="NETCDF3_CLASSIC")
            reloaded = iris.load_cube(filename)
            self.assertCML(
                reloaded,
                ("netcdf", "uint32_auxiliary_coord_netcdf3.cml"),
                checksum=False,
            )

    def test_uint32_data_netcdf3(self):
        self.cube.data = self.cube.data.astype(np.uint32)
        with self.temp_filename(suffix=".nc") as filename:
            iris.save(self.cube, filename, netcdf_format="NETCDF3_CLASSIC")
            reloaded = iris.load_cube(filename)
            self.assertCML(reloaded, ("netcdf", "uint32_data_netcdf3.cml"))

    def test_uint64_dimension_coord_netcdf3(self):
        # Points that cannot be safely cast to int32.
        coord = iris.coords.DimCoord(
            np.array([0, 18446744073709551615], dtype=np.uint64), long_name="x"
        )
        self.cube.add_dim_coord(coord, 0)
        with self.temp_filename(suffix=".nc") as filename:
            with self.assertRaises(ValueError):
                iris.save(self.cube, filename, netcdf_format="NETCDF3_CLASSIC")

    def test_uint64_auxiliary_coord_netcdf3(self):
        # Points that cannot be safely cast to int32.
        coord = iris.coords.AuxCoord(
            np.array([0, 18446744073709551615], dtype=np.uint64), long_name="x"
        )
        self.cube.add_aux_coord(coord, 0)
        with self.temp_filename(suffix=".nc") as filename:
            with self.assertRaises(ValueError):
                iris.save(self.cube, filename, netcdf_format="NETCDF3_CLASSIC")

    def test_uint64_data_netcdf3(self):
        # Data that cannot be safely cast to int32.
        self.cube.data = self.cube.data.astype(np.uint64)
        self.cube.data[0, 1] = 18446744073709551615
        with self.temp_filename(suffix=".nc") as filename:
            with self.assertRaises(ValueError):
                iris.save(self.cube, filename, netcdf_format="NETCDF3_CLASSIC")


class TestCFStandardName(tests.IrisTest):
    def setUp(self):
        pass

    def test_std_name_lookup_pass(self):
        # Test performing a CF standard name look-up hit.
        self.assertTrue("time" in iris.std_names.STD_NAMES)

    def test_std_name_lookup_fail(self):
        # Test performing a CF standard name look-up miss.
        self.assertFalse("phenomenon_time" in iris.std_names.STD_NAMES)


@tests.skip_data
class TestNetCDFUKmoProcessFlags(tests.IrisTest):
    def test_process_flags(self):
        # Test single process flags
        for _, process_desc in iris.fileformats.pp.LBPROC_PAIRS[1:]:
            # Get basic cube and set process flag manually
            ll_cube = stock.lat_lon_cube()
            ll_cube.attributes["ukmo__process_flags"] = (process_desc,)

            # Save cube to netCDF
            with self.temp_filename(suffix=".nc") as temp_filename:
                iris.save(ll_cube, temp_filename)

                # Reload cube
                cube = iris.load_cube(temp_filename)

                # Check correct number and type of flags
                self.assertTrue(
                    len(cube.attributes["ukmo__process_flags"]) == 1,
                    "Mismatch in number of process flags.",
                )
                process_flag = cube.attributes["ukmo__process_flags"][0]
                self.assertEqual(process_flag, process_desc)

        # Test mutiple process flags
        multiple_bit_values = ((128, 64), (4096, 1024), (8192, 1024))

        # Maps lbproc value to the process flags that should be created
        multiple_map = {
            bits: [iris.fileformats.pp.lbproc_map[bit] for bit in bits]
            for bits in multiple_bit_values
        }

        for bits, descriptions in multiple_map.items():

            ll_cube = stock.lat_lon_cube()
            ll_cube.attributes["ukmo__process_flags"] = descriptions

            # Save cube to netCDF
            with self.temp_filename(suffix=".nc") as temp_filename:
                iris.save(ll_cube, temp_filename)

                # Reload cube
                cube = iris.load_cube(temp_filename)

                # Check correct number and type of flags
                process_flags = cube.attributes["ukmo__process_flags"]
                self.assertTrue(
                    len(process_flags) == len(bits),
                    "Mismatch in " "number of process flags.",
                )
                self.assertEqual(set(process_flags), set(descriptions))


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the :class:`iris.coords.Cell` class."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import datetime
from unittest import mock

import numpy as np
import cftime

from iris.coords import Cell
from iris.time import PartialDateTime


class Test___common_cmp__(tests.IrisTest):
    def assert_raises_on_comparison(self, cell, other, exception_type, regexp):
        with self.assertRaisesRegex(exception_type, regexp):
            cell < other
        with self.assertRaisesRegex(exception_type, regexp):
            cell <= other
        with self.assertRaisesRegex(exception_type, regexp):
            cell > other
        with self.assertRaisesRegex(exception_type, regexp):
            cell >= other

    def test_cftime_cell(self):
        # Check that cell comparison when the cell contains
        # cftime.datetime objects raises an exception otherwise
        # this will fall back to id comparison producing unreliable
        # results.
        cell = Cell(cftime.datetime(2010, 3, 21))
        dt = mock.Mock(timetuple=mock.Mock())
        self.assert_raises_on_comparison(
            cell, dt, TypeError, "determine the order of cftime"
        )
        self.assert_raises_on_comparison(
            cell, 23, TypeError, "determine the order of cftime"
        )
        self.assert_raises_on_comparison(
            cell, "hello", TypeError, "Unexpected type.*str"
        )

    def test_cftime_other(self):
        # Check that cell comparison to a cftime.datetime object
        # raises an exception otherwise this will fall back to id comparison
        # producing unreliable results.
        dt = cftime.datetime(2010, 3, 21)
        cell = Cell(mock.Mock(timetuple=mock.Mock()))
        self.assert_raises_on_comparison(
            cell, dt, TypeError, "determine the order of cftime"
        )

    def test_PartialDateTime_bounded_cell(self):
        # Check that bounded comparisions to a PartialDateTime
        # raise an exception. These are not supported as they
        # depend on the calendar.
        dt = PartialDateTime(month=6)
        cell = Cell(
            datetime.datetime(2010, 1, 1),
            bound=[
                datetime.datetime(2010, 1, 1),
                datetime.datetime(2011, 1, 1),
            ],
        )
        self.assert_raises_on_comparison(
            cell, dt, TypeError, "bounded region for datetime"
        )

    def test_PartialDateTime_unbounded_cell(self):
        # Check that cell comparison works with PartialDateTimes.
        dt = PartialDateTime(month=6)
        cell = Cell(cftime.datetime(2010, 3, 1))
        self.assertLess(cell, dt)
        self.assertGreater(dt, cell)
        self.assertLessEqual(cell, dt)
        self.assertGreaterEqual(dt, cell)

    def test_datetime_unbounded_cell(self):
        # Check that cell comparison works with datetimes.
        dt = datetime.datetime(2000, 6, 15)
        cell = Cell(datetime.datetime(2000, 1, 1))
        # Note the absence of the inverse of these
        # e.g. self.assertGreater(dt, cell).
        # See http://bugs.python.org/issue8005
        self.assertLess(cell, dt)
        self.assertLessEqual(cell, dt)


class Test___eq__(tests.IrisTest):
    def test_datetimelike(self):
        # Check that cell equality works with objects with a "timetuple".
        dt = mock.Mock(timetuple=mock.Mock())
        cell = mock.MagicMock(
            spec=Cell, point=datetime.datetime(2010, 3, 21), bound=None
        )
        _ = cell == dt
        cell.__eq__.assert_called_once_with(dt)

    def test_datetimelike_bounded_cell(self):
        # Check that equality with a datetime-like bounded cell
        # raises an error. This is not supported as it
        # depends on the calendar which is not always known from
        # the datetime-like bound objects.
        other = mock.Mock(timetuple=mock.Mock())
        cell = Cell(
            point=object(),
            bound=[
                mock.Mock(timetuple=mock.Mock()),
                mock.Mock(timetuple=mock.Mock()),
            ],
        )
        with self.assertRaisesRegex(TypeError, "bounded region for datetime"):
            cell == other

    def test_PartialDateTime_other(self):
        cell = Cell(datetime.datetime(2010, 3, 2))
        # A few simple cases.
        self.assertEqual(cell, PartialDateTime(month=3))
        self.assertNotEqual(cell, PartialDateTime(month=3, hour=12))
        self.assertNotEqual(cell, PartialDateTime(month=4))


class Test_contains_point(tests.IrisTest):
    def test_datetimelike_bounded_cell(self):
        point = object()
        cell = Cell(
            point=object(),
            bound=[
                mock.Mock(timetuple=mock.Mock()),
                mock.Mock(timetuple=mock.Mock()),
            ],
        )
        with self.assertRaisesRegex(TypeError, "bounded region for datetime"):
            cell.contains_point(point)

    def test_datetimelike_point(self):
        point = mock.Mock(timetuple=mock.Mock())
        cell = Cell(point=object(), bound=[object(), object()])
        with self.assertRaisesRegex(TypeError, "bounded region for datetime"):
            cell.contains_point(point)


class Test_numpy_comparison(tests.IrisTest):
    """
    Unit tests to check that the results of comparisons with numpy types can be
    used as truth values."""

    def test_cell_lhs(self):
        cell = Cell(point=1.5)
        n = np.float64(1.2)

        try:
            bool(cell < n)
            bool(cell <= n)
            bool(cell > n)
            bool(cell >= n)
            bool(cell == n)
            bool(cell != n)
        except:  # noqa
            self.fail(
                "Result of comparison could not be used as a truth value"
            )

    def test_cell_rhs(self):
        cell = Cell(point=1.5)
        n = np.float64(1.2)

        try:
            bool(n < cell)
            bool(n <= cell)
            bool(n > cell)
            bool(n >= cell)
            bool(n == cell)
            bool(n != cell)
        except:  # noqa
            self.fail(
                "Result of comparison could not be used as a truth value"
            )


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Integration tests for regridding."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np

import iris
from iris.analysis._regrid import RectilinearRegridder as Regridder
from iris.coord_systems import GeogCS
from iris.coords import DimCoord
from iris.cube import Cube
from iris.tests.stock import global_pp, simple_3d
from iris.analysis import UnstructuredNearest


@tests.skip_data
class TestOSGBToLatLon(tests.IrisTest):
    def setUp(self):
        path = tests.get_data_path(
            (
                "NIMROD",
                "uk2km",
                "WO0000000003452",
                "201007020900_u1096_ng_ey00_visibility0180_screen_2km",
            )
        )
        self.src = iris.load_cube(path)[0]
        # Cast up to float64, to work around numpy<=1.8 bug with means of
        # arrays of 32bit floats.
        self.src.data = self.src.data.astype(np.float64)
        self.grid = Cube(np.empty((73, 96)))
        cs = GeogCS(6370000)
        lat = DimCoord(
            np.linspace(46, 65, 73),
            "latitude",
            units="degrees",
            coord_system=cs,
        )
        lon = DimCoord(
            np.linspace(-14, 8, 96),
            "longitude",
            units="degrees",
            coord_system=cs,
        )
        self.grid.add_dim_coord(lat, 0)
        self.grid.add_dim_coord(lon, 1)

    def _regrid(self, method):
        regridder = Regridder(self.src, self.grid, method, "mask")
        result = regridder(self.src)
        return result

    def test_linear(self):
        res = self._regrid("linear")
        self.assertArrayShapeStats(res, (73, 96), 17799.296120, 11207.701323)

    def test_nearest(self):
        res = self._regrid("nearest")
        self.assertArrayShapeStats(res, (73, 96), 17808.068828, 11225.314310)


@tests.skip_data
class TestGlobalSubsample(tests.IrisTest):
    def setUp(self):
        self.src = global_pp()
        _ = self.src.data
        # Cast up to float64, to work around numpy<=1.8 bug with means of
        # arrays of 32bit floats.
        self.src.data = self.src.data.astype(np.float64)
        # Subsample and shift the target grid so that we can see a visual
        # difference between regridding scheme methods.
        grid = self.src[1::2, 1::3]
        grid.coord("latitude").points = grid.coord("latitude").points + 1
        grid.coord("longitude").points = grid.coord("longitude").points + 1
        self.grid = grid

    def _regrid(self, method):
        regridder = Regridder(self.src, self.grid, method, "mask")
        result = regridder(self.src)
        return result

    def test_linear(self):
        res = self._regrid("linear")
        self.assertArrayShapeStats(res, (36, 32), 280.35907, 15.997223)

    def test_nearest(self):
        res = self._regrid("nearest")
        self.assertArrayShapeStats(res, (36, 32), 280.33726, 16.064001)


@tests.skip_data
class TestUnstructured(tests.IrisTest):
    def setUp(self):
        path = tests.get_data_path(
            ("NetCDF", "unstructured_grid", "theta_nodal_xios.nc")
        )
        self.src = iris.load_cube(path, "Potential Temperature")
        self.grid = simple_3d()[0, :, :]

    def test_nearest(self):
        res = self.src.regrid(self.grid, UnstructuredNearest())
        self.assertArrayShapeStats(res, (1, 6, 3, 4), 315.890808, 11.000724)


class TestZonalMean_global(tests.IrisTest):
    def setUp(self):
        np.random.seed(0)
        self.src = iris.cube.Cube(np.random.random_integers(0, 10, (140, 1)))
        s_crs = iris.coord_systems.GeogCS(6371229.0)
        sy_coord = iris.coords.DimCoord(
            np.linspace(-90, 90, 140),
            standard_name="latitude",
            units="degrees",
            coord_system=s_crs,
        )
        sx_coord = iris.coords.DimCoord(
            -180,
            bounds=[-180, 180],
            standard_name="longitude",
            units="degrees",
            circular=True,
            coord_system=s_crs,
        )
        self.src.add_dim_coord(sy_coord, 0)
        self.src.add_dim_coord(sx_coord, 1)

    def test_linear_same_crs_global(self):
        # Regrid the zonal mean onto an identical coordinate system target, but
        # on a different set of longitudes - which should result in no change.
        points = [-150, -90, -30, 30, 90, 150]
        bounds = [
            [-180, -120],
            [-120, -60],
            [-60, 0],
            [0, 60],
            [60, 120],
            [120, 180],
        ]
        sx_coord = self.src.coord(axis="x")
        sy_coord = self.src.coord(axis="y")
        x_coord = sx_coord.copy(points, bounds=bounds)
        grid = iris.cube.Cube(
            np.zeros([sy_coord.points.size, x_coord.points.size])
        )
        grid.add_dim_coord(sy_coord, 0)
        grid.add_dim_coord(x_coord, 1)

        res = self.src.regrid(grid, iris.analysis.Linear())

        # Ensure data remains unchanged.
        # (the same along each column)
        self.assertTrue(
            np.array(
                [
                    (res.data[:, 0] - res.data[:, i]).max()
                    for i in range(1, res.shape[1])
                ]
            ).max()
            < 1e-10
        )
        self.assertArrayAlmostEqual(res.data[:, 0], self.src.data.reshape(-1))


class TestZonalMean_regional(TestZonalMean_global, tests.IrisTest):
    def setUp(self):
        super().setUp()

        # Define a target grid and a target result (what we expect the
        # regridder to return).
        sx_coord = self.src.coord(axis="x")
        sy_coord = self.src.coord(axis="y")
        grid_crs = iris.coord_systems.RotatedGeogCS(
            37.5, 177.5, ellipsoid=iris.coord_systems.GeogCS(6371229.0)
        )
        grid_x = sx_coord.copy(np.linspace(350, 370, 100))
        grid_x.circular = False
        grid_x.coord_system = grid_crs
        grid_y = sy_coord.copy(np.linspace(-10, 10, 100))
        grid_y.coord_system = grid_crs
        grid = iris.cube.Cube(
            np.zeros([grid_y.points.size, grid_x.points.size])
        )
        grid.add_dim_coord(grid_y, 0)
        grid.add_dim_coord(grid_x, 1)

        # The target result is derived by regridding a multi-column version of
        # the source to the target (i.e. turning a zonal mean regrid into a
        # conventional regrid).
        self.tar = self.zonal_mean_as_multi_column(self.src).regrid(
            grid, iris.analysis.Linear()
        )
        self.grid = grid

    def zonal_mean_as_multi_column(self, src_cube):
        # Munge the source (duplicate source latitudes) so that we can
        # utilise linear regridding as a conventional problem (that is, to
        # duplicate columns so that it is no longer a zonal mean problem).
        src_cube2 = src_cube.copy()
        src_cube2.coord(axis="x").points = -90
        src_cube2.coord(axis="x").bounds = [-180, 0]
        src_cube.coord(axis="x").points = 90
        src_cube.coord(axis="x").bounds = [0, 180]
        src_cubes = iris.cube.CubeList([src_cube, src_cube2])
        return src_cubes.concatenate_cube()

    def test_linear_rotated_regional(self):
        # Ensure that zonal mean source data is linearly interpolated onto a
        # high resolution target.
        regridder = iris.analysis.Linear()
        res = self.src.regrid(self.grid, regridder)
        self.assertArrayAlmostEqual(res.data, self.tar.data)

    def test_linear_rotated_regional_no_extrapolation(self):
        # Capture the case where our source remains circular but we don't use
        # extrapolation.
        regridder = iris.analysis.Linear(extrapolation_mode="nan")
        res = self.src.regrid(self.grid, regridder)
        self.assertArrayAlmostEqual(res.data, self.tar.data)

    def test_linear_rotated_regional_not_circular(self):
        # Capture the case where our source is not circular but we utilise
        # extrapolation.
        regridder = iris.analysis.Linear()
        self.src.coord(axis="x").circular = False
        res = self.src.regrid(self.grid, regridder)
        self.assertArrayAlmostEqual(res.data, self.tar.data)

    def test_linear_rotated_regional_no_extrapolation_not_circular(self):
        # Confirm how zonal mean actually works in so far as, that
        # extrapolation and circular source handling is the means by which
        # these usecases are supported.
        # In the case where the source is neither using extrapolation and is
        # not circular, then 'nan' values will result (as we would expect).
        regridder = iris.analysis.Linear(extrapolation_mode="nan")
        self.src.coord(axis="x").circular = False
        res = self.src.regrid(self.grid, regridder)
        self.assertTrue(np.isnan(res.data).all())


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test the fast loading of structured Fieldsfiles.

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests
from iris.cube import CubeList

from iris.fileformats.um import load_cubes as load


@tests.skip_data
class TestStructuredLoadFF(tests.IrisTest):
    def setUp(self):
        self.fname = tests.get_data_path(("FF", "structured", "small"))

    def _merge_cubes(self, cubes):
        # Merge the 2D cubes returned by `iris.fileformats.um.load_cubes`.
        return CubeList(cubes).merge_cube()

    def test_simple(self):
        list_of_cubes = list(load(self.fname, None))
        cube = self._merge_cubes(list_of_cubes)
        self.assertCML(cube)

    def test_simple_callback(self):
        def callback(cube, field, filename):
            cube.attributes["processing"] = "fast-ff"

        list_of_cubes = list(load(self.fname, callback=callback))
        cube = self._merge_cubes(list_of_cubes)
        self.assertCML(cube)


if __name__ == "__main__":
    tests.main()



"""
Cross Section Plots
===================

This example demonstrates contour plots of a cross-sectioned multi-dimensional
cube which features a hybrid height vertical coordinate system.

"""

import matplotlib.pyplot as plt

import iris
import iris.plot as iplt
import iris.quickplot as qplt


def main():
    # Load some test data.
    fname = iris.sample_data_path("hybrid_height.nc")
    theta = iris.load_cube(fname, "air_potential_temperature")

    # Extract a single height vs longitude cross-section. N.B. This could
    # easily be changed to extract a specific slice, or even to loop over *all*
    # cross section slices.
    cross_section = next(
        theta.slices(["grid_longitude", "model_level_number"])
    )

    qplt.contourf(
        cross_section, coords=["grid_longitude", "altitude"], cmap="RdBu_r"
    )
    iplt.show()

    # Now do the equivalent plot, only against model level
    plt.figure()

    qplt.contourf(
        cross_section,
        coords=["grid_longitude", "model_level_number"],
        cmap="RdBu_r",
    )
    iplt.show()


if __name__ == "__main__":
    main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
A package for handling multi-dimensional data and associated metadata.

.. note ::

    The Iris documentation has further usage information, including
    a :ref:`user guide <user_guide_index>` which should be the first port of
    call for new users.

The functions in this module provide the main way to load and/or save
your data.

The :func:`load` function provides a simple way to explore data from
the interactive Python prompt. It will convert the source data into
:class:`Cubes <iris.cube.Cube>`, and combine those cubes into
higher-dimensional cubes where possible.

The :func:`load_cube` and :func:`load_cubes` functions are similar to
:func:`load`, but they raise an exception if the number of cubes is not
what was expected. They are more useful in scripts, where they can
provide an early sanity check on incoming data.

The :func:`load_raw` function is provided for those occasions where the
automatic combination of cubes into higher-dimensional cubes is
undesirable. However, it is intended as a tool of last resort! If you
experience a problem with the automatic combination process then please
raise an issue with the Iris developers.

To persist a cube to the file-system, use the :func:`save` function.

All the load functions share very similar arguments:

    * uris:
        Either a single filename/URI expressed as a string, or an
        iterable of filenames/URIs.

        Filenames can contain `~` or `~user` abbreviations, and/or
        Unix shell-style wildcards (e.g. `*` and `?`). See the
        standard library function :func:`os.path.expanduser` and
        module :mod:`fnmatch` for more details.

    * constraints:
        Either a single constraint, or an iterable of constraints.
        Each constraint can be either a string, an instance of
        :class:`iris.Constraint`, or an instance of
        :class:`iris.AttributeConstraint`.  If the constraint is a string
        it will be used to match against cube.name().

        .. _constraint_egs:

        For example::

            # Load air temperature data.
            load_cube(uri, 'air_temperature')

            # Load data with a specific model level number.
            load_cube(uri, iris.Constraint(model_level_number=1))

            # Load data with a specific STASH code.
            load_cube(uri, iris.AttributeConstraint(STASH='m01s00i004'))

    * callback:
        A function to add metadata from the originating field and/or URI which
        obeys the following rules:

        1. Function signature must be: ``(cube, field, filename)``.
        2. Modifies the given cube inplace, unless a new cube is
           returned by the function.
        3. If the cube is to be rejected the callback must raise
           an :class:`iris.exceptions.IgnoreCubeException`.

        For example::

            def callback(cube, field, filename):
                # Extract ID from filenames given as: <prefix>__<exp_id>
                experiment_id = filename.split('__')[1]
                experiment_coord = iris.coords.AuxCoord(
                    experiment_id, long_name='experiment_id')
                cube.add_aux_coord(experiment_coord)

"""

import contextlib
import glob
import itertools
import os.path
import threading

import iris.config
import iris.cube
import iris._constraints
from iris._deprecation import IrisDeprecation, warn_deprecated
import iris.fileformats
import iris.io


try:
    import iris_sample_data
except ImportError:
    iris_sample_data = None


# Iris revision.
__version__ = "3.0.2"

# Restrict the names imported when using "from iris import *"
__all__ = [
    "load",
    "load_cube",
    "load_cubes",
    "load_raw",
    "save",
    "Constraint",
    "AttributeConstraint",
    "NameConstraint",
    "sample_data_path",
    "site_configuration",
    "Future",
    "FUTURE",
    "IrisDeprecation",
]


Constraint = iris._constraints.Constraint
AttributeConstraint = iris._constraints.AttributeConstraint
NameConstraint = iris._constraints.NameConstraint


class Future(threading.local):
    """Run-time configuration controller."""

    def __init__(self):
        """
        A container for run-time options controls.

        To adjust the values simply update the relevant attribute from
        within your code. For example::

            iris.FUTURE.example_future_flag = False

        If Iris code is executed with multiple threads, note the values of
        these options are thread-specific.

        .. note::

            iris.FUTURE.example_future_flag does not exist. It is provided
            as an example because there are currently no flags in
            iris.Future.

        """
        # The flag 'example_future_flag' is provided as a future reference
        # for the structure of this class.
        #
        # self.__dict__['example_future_flag'] = example_future_flag
        pass

    def __repr__(self):

        # msg = ('Future(example_future_flag={})')
        # return msg.format(self.example_future_flag)
        msg = "Future()"
        return msg.format()

    # deprecated_options = {'example_future_flag': 'warning',}
    deprecated_options = {}

    def __setattr__(self, name, value):
        if name in self.deprecated_options:
            level = self.deprecated_options[name]
            if level == "error" and not value:
                emsg = (
                    "setting the 'Future' property {prop!r} has been "
                    "deprecated to be removed in a future release, and "
                    "deprecated {prop!r} behaviour has been removed. "
                    "Please remove code that sets this property."
                )
                raise AttributeError(emsg.format(prop=name))
            else:
                msg = (
                    "setting the 'Future' property {!r} is deprecated "
                    "and will be removed in a future release. "
                    "Please remove code that sets this property."
                )
                warn_deprecated(msg.format(name))
        if name not in self.__dict__:
            msg = "'Future' object has no attribute {!r}".format(name)
            raise AttributeError(msg)
        self.__dict__[name] = value

    @contextlib.contextmanager
    def context(self, **kwargs):
        """
        Return a context manager which allows temporary modification of
        the option values for the active thread.

        On entry to the `with` statement, all keyword arguments are
        applied to the Future object. On exit from the `with`
        statement, the previous state is restored.

        For example::
            with iris.FUTURE.context(example_future_flag=False):
                # ... code that expects some past behaviour

        .. note::

            iris.FUTURE.example_future_flag does not exist and is
            provided only as an example since there are currently no
            flags in Future.

        """
        # Save the current context
        current_state = self.__dict__.copy()
        # Update the state
        for name, value in kwargs.items():
            setattr(self, name, value)
        try:
            yield
        finally:
            # Return the state
            self.__dict__.clear()
            self.__dict__.update(current_state)


#: Object containing all the Iris run-time options.
FUTURE = Future()


# Initialise the site configuration dictionary.
#: Iris site configuration dictionary.
site_configuration = {}

try:
    from iris.site_config import update as _update
except ImportError:
    pass
else:
    _update(site_configuration)


def _generate_cubes(uris, callback, constraints):
    """Returns a generator of cubes given the URIs and a callback."""
    if isinstance(uris, str):
        uris = [uris]

    # Group collections of uris by their iris handler
    # Create list of tuples relating schemes to part names
    uri_tuples = sorted(iris.io.decode_uri(uri) for uri in uris)

    for scheme, groups in itertools.groupby(uri_tuples, key=lambda x: x[0]):
        # Call each scheme handler with the appropriate URIs
        if scheme == "file":
            part_names = [x[1] for x in groups]
            for cube in iris.io.load_files(part_names, callback, constraints):
                yield cube
        elif scheme in ["http", "https"]:
            urls = [":".join(x) for x in groups]
            for cube in iris.io.load_http(urls, callback):
                yield cube
        else:
            raise ValueError("Iris cannot handle the URI scheme: %s" % scheme)


def _load_collection(uris, constraints=None, callback=None):
    try:
        cubes = _generate_cubes(uris, callback, constraints)
        result = iris.cube._CubeFilterCollection.from_cubes(cubes, constraints)
    except EOFError as e:
        raise iris.exceptions.TranslationError(
            "The file appears empty or incomplete: {!r}".format(str(e))
        )
    return result


def load(uris, constraints=None, callback=None):
    """
    Loads any number of Cubes for each constraint.

    For a full description of the arguments, please see the module
    documentation for :mod:`iris`.

    Args:

    * uris:
        One or more filenames/URIs.

    Kwargs:

    * constraints:
        One or more constraints.
    * callback:
        A modifier/filter function.

    Returns:
        An :class:`iris.cube.CubeList`. Note that there is no inherent order
        to this :class:`iris.cube.CubeList` and it should be treated as if it
        were random.

    """
    return _load_collection(uris, constraints, callback).merged().cubes()


def load_cube(uris, constraint=None, callback=None):
    """
    Loads a single cube.

    For a full description of the arguments, please see the module
    documentation for :mod:`iris`.

    Args:

    * uris:
        One or more filenames/URIs.

    Kwargs:

    * constraints:
        A constraint.
    * callback:
        A modifier/filter function.

    Returns:
        An :class:`iris.cube.Cube`.

    """
    constraints = iris._constraints.list_of_constraints(constraint)
    if len(constraints) != 1:
        raise ValueError("only a single constraint is allowed")

    cubes = _load_collection(uris, constraints, callback).cubes()

    try:
        cube = cubes.merge_cube()
    except iris.exceptions.MergeError as e:
        raise iris.exceptions.ConstraintMismatchError(str(e))
    except ValueError:
        raise iris.exceptions.ConstraintMismatchError("no cubes found")

    return cube


def load_cubes(uris, constraints=None, callback=None):
    """
    Loads exactly one Cube for each constraint.

    For a full description of the arguments, please see the module
    documentation for :mod:`iris`.

    Args:

    * uris:
        One or more filenames/URIs.

    Kwargs:

    * constraints:
        One or more constraints.
    * callback:
        A modifier/filter function.

    Returns:
        An :class:`iris.cube.CubeList`. Note that there is no inherent order
        to this :class:`iris.cube.CubeList` and it should be treated as if it
        were random.

    """
    # Merge the incoming cubes
    collection = _load_collection(uris, constraints, callback).merged()

    # Make sure we have exactly one merged cube per constraint
    bad_pairs = [pair for pair in collection.pairs if len(pair) != 1]
    if bad_pairs:
        fmt = "   {} -> {} cubes"
        bits = [fmt.format(pair.constraint, len(pair)) for pair in bad_pairs]
        msg = "\n" + "\n".join(bits)
        raise iris.exceptions.ConstraintMismatchError(msg)

    return collection.cubes()


def load_raw(uris, constraints=None, callback=None):
    """
    Loads non-merged cubes.

    This function is provided for those occasions where the automatic
    combination of cubes into higher-dimensional cubes is undesirable.
    However, it is intended as a tool of last resort! If you experience
    a problem with the automatic combination process then please raise
    an issue with the Iris developers.

    For a full description of the arguments, please see the module
    documentation for :mod:`iris`.

    Args:

    * uris:
        One or more filenames/URIs.

    Kwargs:

    * constraints:
        One or more constraints.
    * callback:
        A modifier/filter function.

    Returns:
        An :class:`iris.cube.CubeList`.

    """
    from iris.fileformats.um._fast_load import _raw_structured_loading

    with _raw_structured_loading():
        return _load_collection(uris, constraints, callback).cubes()


save = iris.io.save


def sample_data_path(*path_to_join):
    """
    Given the sample data resource, returns the full path to the file.

    .. note::

        This function is only for locating files in the iris sample data
        collection (installed separately from iris). It is not needed or
        appropriate for general file access.

    """
    target = os.path.join(*path_to_join)
    if os.path.isabs(target):
        raise ValueError(
            "Absolute paths, such as {!r}, are not supported.\n"
            "NB. This function is only for locating files in the "
            "iris sample data collection. It is not needed or "
            "appropriate for general file access.".format(target)
        )
    if iris_sample_data is not None:
        target = os.path.join(iris_sample_data.path, target)
    else:
        raise ImportError(
            "Please install the 'iris-sample-data' package to "
            "access sample data."
        )
    if not glob.glob(target):
        raise ValueError(
            "Sample data file(s) at {!r} not found.\n"
            "NB. This function is only for locating files in the "
            "iris sample data collection. It is not needed or "
            "appropriate for general file access.".format(target)
        )
    return target



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

# import iris tests first so that some things can be initialised before importing anything else
import iris.tests as tests

import operator
import math

import cf_units
import numpy as np
import numpy.ma as ma

import iris
import iris.analysis.maths
from iris.coords import AuxCoord, DimCoord
from iris.cube import Cube
from iris.exceptions import NotYetImplementedError
import iris.tests.stock


@tests.skip_data
class TestBasicMaths(tests.IrisTest):
    def setUp(self):
        self.cube = iris.tests.stock.global_pp()
        self.cube.data = self.cube.data - 260

    def test_abs(self):
        a = self.cube

        b = iris.analysis.maths.abs(a, in_place=False)
        self.assertCML(a, ("analysis", "maths_original.cml"))
        self.assertCML(b, ("analysis", "abs.cml"))

        iris.analysis.maths.abs(a, in_place=True)
        self.assertCML(b, ("analysis", "abs.cml"))
        self.assertCML(a, ("analysis", "abs.cml"))

    def test_minus(self):
        a = self.cube
        e = self.cube.copy()

        # Check that the subtraction has had no effect on the original
        self.assertCML(a, ("analysis", "maths_original.cml"))

        d = a - a
        self.assertCML(d, ("analysis", "subtract.cml"))

        # Check that the subtraction has had no effect on the original
        self.assertCML(a, ("analysis", "maths_original.cml"))

        c = iris.analysis.maths.subtract(e, e)
        self.assertCML(c, ("analysis", "subtract.cml"))

        # Check that the subtraction has had no effect on the original
        self.assertCML(e, ("analysis", "maths_original.cml"))

    def test_minus_with_data_describing_coordinate(self):
        a = self.cube
        e = self.cube.copy()
        lat = e.coord("latitude")
        lat.points = lat.points + 100

        # Cannot ignore a axis describing coordinate
        self.assertRaises(ValueError, iris.analysis.maths.subtract, a, e)

    def test_minus_scalar(self):
        a = self.cube

        self.assertCML(a, ("analysis", "maths_original.cml"))

        b = a - 200
        self.assertCML(b, ("analysis", "subtract_scalar.cml"))
        # Check that the subtraction has had no effect on the original
        self.assertCML(a, ("analysis", "maths_original.cml"))

    def test_minus_array(self):
        a = self.cube
        data_array = self.cube.copy().data

        # check that the file has not changed (avoids false positives by failing early)
        self.assertCML(a, ("analysis", "maths_original.cml"))

        # subtract an array of exactly the same shape as the original
        b = a - data_array
        self.assertArrayEqual(b.data, np.array(0, dtype=np.float32))
        self.assertCML(b, ("analysis", "subtract_array.cml"), checksum=False)

        # subtract an array of the same number of dimensions, but with one of the dimensions having len 1
        b = a - data_array[:, 0:1]
        self.assertArrayEqual(b.data[:, 0:1], np.array(0, dtype=np.float32))
        self.assertArrayEqual(b.data[:, 1:2], b.data[:, 1:2])

        # subtract an array of 1 dimension fewer than the cube
        b = a - data_array[0, :]
        self.assertArrayEqual(b.data[0, :], np.array(0, dtype=np.float32))
        self.assertArrayEqual(b.data[:, 1:2], b.data[:, 1:2])

        # subtract an array of 1 dimension more than the cube
        d_array = data_array.reshape(
            data_array.shape[0], data_array.shape[1], 1
        )
        self.assertRaises(ValueError, iris.analysis.maths.subtract, a, d_array)

        # Check that the subtraction has had no effect on the original
        self.assertCML(a, ("analysis", "maths_original.cml"))

    def test_minus_coord(self):
        a = self.cube

        xdim = a.ndim - 1
        ydim = a.ndim - 2
        c_x = DimCoord(
            points=np.arange(a.shape[xdim]),
            long_name="x_coord",
            units=self.cube.units,
        )
        c_y = AuxCoord(
            points=np.arange(a.shape[ydim]),
            long_name="y_coord",
            units=self.cube.units,
        )

        self.assertCML(a, ("analysis", "maths_original.cml"))

        b = iris.analysis.maths.subtract(a, c_x, dim=1)
        self.assertCML(b, ("analysis", "subtract_coord_x.cml"))
        # Check that the subtraction has had no effect on the original
        self.assertCML(a, ("analysis", "maths_original.cml"))

        b = iris.analysis.maths.subtract(a, c_y, dim=0)
        self.assertCML(b, ("analysis", "subtract_coord_y.cml"))
        # Check that the subtraction has had no effect on the original
        self.assertCML(a, ("analysis", "maths_original.cml"))

    def test_addition_scalar(self):
        a = self.cube

        self.assertCML(a, ("analysis", "maths_original.cml"))

        b = a + 200
        self.assertCML(b, ("analysis", "addition_scalar.cml"))
        # Check that the addition has had no effect on the original
        self.assertCML(a, ("analysis", "maths_original.cml"))

    def test_addition_coord(self):
        a = self.cube

        xdim = a.ndim - 1
        ydim = a.ndim - 2
        c_x = DimCoord(
            points=np.arange(a.shape[xdim]),
            long_name="x_coord",
            units=self.cube.units,
        )
        c_y = AuxCoord(
            points=np.arange(a.shape[ydim]),
            long_name="y_coord",
            units=self.cube.units,
        )

        self.assertCML(a, ("analysis", "maths_original.cml"))

        b = iris.analysis.maths.add(a, c_x, dim=1)
        self.assertCML(b, ("analysis", "addition_coord_x.cml"))
        # Check that the addition has had no effect on the original
        self.assertCML(a, ("analysis", "maths_original.cml"))

        b = iris.analysis.maths.add(a, c_y, dim=0)
        self.assertCML(b, ("analysis", "addition_coord_y.cml"))
        # Check that the addition has had no effect on the original
        self.assertCML(a, ("analysis", "maths_original.cml"))

    def test_addition(self):
        a = self.cube

        c = a + a
        self.assertCML(c, ("analysis", "addition.cml"))
        # Check that the addition has had no effect on the original
        self.assertCML(a, ("analysis", "maths_original.cml"))

    def test_addition_different_standard_name(self):
        a = self.cube.copy()
        b = self.cube.copy()
        b.rename("my cube data")
        c = a + b
        self.assertCML(
            c, ("analysis", "addition_different_std_name.cml"), checksum=False
        )

    def test_addition_fail(self):
        a = self.cube

        xdim = a.ndim - 1
        ydim = a.ndim - 2
        c_axis_length_fail = DimCoord(
            points=np.arange(a.shape[ydim]),
            long_name="x_coord",
            units=self.cube.units,
        )
        c_unit_fail = AuxCoord(
            points=np.arange(a.shape[xdim]), long_name="x_coord", units="volts"
        )

        self.assertRaises(
            ValueError, iris.analysis.maths.add, a, c_axis_length_fail
        )
        self.assertRaises(
            NotYetImplementedError,
            iris.analysis.maths.add,
            a,
            c_unit_fail,
        )

    def test_addition_in_place(self):
        a = self.cube

        b = iris.analysis.maths.add(a, self.cube, in_place=True)
        self.assertTrue(b is a)
        self.assertCML(a, ("analysis", "addition_in_place.cml"))

    def test_addition_in_place_coord(self):
        a = self.cube

        # scalar is promoted to a coordinate internally
        b = iris.analysis.maths.add(a, 1000, in_place=True)
        self.assertTrue(b is a)
        self.assertCML(a, ("analysis", "addition_in_place_coord.cml"))

    def test_addition_different_attributes(self):
        a = self.cube.copy()
        b = self.cube.copy()
        b.attributes["my attribute"] = "foobar"
        c = a + b
        self.assertIsNone(c.standard_name)
        expected = {
            "my attribute": "foobar",
            "source": "Data from Met Office Unified Model",
        }
        self.assertEqual(expected, c.attributes)

    def test_apply_ufunc(self):
        a = self.cube

        b = iris.analysis.maths.apply_ufunc(
            np.square,
            a,
            new_name="squared temperature",
            new_unit=a.units ** 2,
            in_place=False,
        )
        self.assertCMLApproxData(a, ("analysis", "apply_ufunc_original.cml"))
        self.assertCMLApproxData(b, ("analysis", "apply_ufunc.cml"))

        b = iris.analysis.maths.apply_ufunc(
            np.square,
            a,
            new_name="squared temperature",
            new_unit=a.units ** 2,
            in_place=True,
        )
        self.assertCMLApproxData(b, ("analysis", "apply_ufunc.cml"))
        self.assertCMLApproxData(a, ("analysis", "apply_ufunc.cml"))

        def vec_mag(u, v):
            return math.sqrt(u ** 2 + v ** 2)

        c = a.copy() + 2

        vec_mag_ufunc = np.frompyfunc(vec_mag, 2, 1)
        b = iris.analysis.maths.apply_ufunc(vec_mag_ufunc, a, c)
        self.assertCMLApproxData(b, ("analysis", "apply_ufunc_frompyfunc.cml"))

    def test_apply_ufunc_fail(self):
        a = self.cube

        # should fail because 'blah' is a string, not a np.ufunc
        self.assertRaises(
            TypeError, iris.analysis.maths.apply_ufunc, "blah", a
        )

        # should fail because math.sqrt is not a np.ufunc
        self.assertRaises(
            TypeError, iris.analysis.maths.apply_ufunc, math.sqrt, a
        )

        # should fail because np.frexp gives 2 arrays as output
        self.assertRaises(
            ValueError, iris.analysis.maths.apply_ufunc, np.frexp, a
        )

    def test_ifunc(self):
        a = self.cube

        my_ifunc = iris.analysis.maths.IFunc(np.square, lambda a: a.units ** 2)
        b = my_ifunc(a, new_name="squared temperature", in_place=False)

        self.assertCMLApproxData(a, ("analysis", "apply_ifunc_original.cml"))
        self.assertCMLApproxData(b, ("analysis", "apply_ifunc.cml"))

        b = my_ifunc(a, new_name="squared temperature", in_place=True)

        self.assertCMLApproxData(b, ("analysis", "apply_ifunc.cml"))
        self.assertCMLApproxData(a, ("analysis", "apply_ifunc.cml"))

        def vec_mag(u, v):
            return math.sqrt(u ** 2 + v ** 2)

        c = a.copy() + 2

        vec_mag_ufunc = np.frompyfunc(vec_mag, 2, 1)
        my_ifunc = iris.analysis.maths.IFunc(
            vec_mag_ufunc, lambda a, b: (a + b).units
        )

        b = my_ifunc(a, c)
        self.assertCMLApproxData(b, ("analysis", "apply_ifunc_frompyfunc.cml"))

    def test_ifunc_init_fail(self):
        # should fail because 'blah' is a string not a python function
        self.assertRaises(
            TypeError,
            iris.analysis.maths.IFunc,
            "blah",
            lambda cube: cf_units.Unit("1"),
        )

        # should fail because np.frexp gives 2 arrays as output
        self.assertRaises(
            ValueError,
            iris.analysis.maths.IFunc,
            np.frexp,
            lambda cube: cf_units.Unit("1"),
        )

        # should fail because data function has 3 arguments
        self.assertRaises(
            ValueError,
            iris.analysis.maths.IFunc,
            lambda a, b, c: a + b + c,
            lambda cube: cf_units.Unit("1"),
        )

    def test_ifunc_call_fail(self):
        a = self.cube

        my_ifunc = iris.analysis.maths.IFunc(np.square, lambda a: a.units ** 2)

        # should now NOT fail because giving 2 arguments to an ifunc that
        # expects only one will now ignore the surplus argument and raise
        # a logging message instead, and go on to perform the operation.
        emsg = "ValueError not raised"
        with self.assertRaisesRegex(AssertionError, emsg):
            with self.assertRaises(ValueError):
                my_ifunc(a, a)

        my_ifunc = iris.analysis.maths.IFunc(
            np.multiply, lambda a: cf_units.Unit("1")
        )

        # should fail because giving 1 arguments to an ifunc that expects
        # 2
        with self.assertRaises(ValueError):
            my_ifunc(a)

        my_ifunc = iris.analysis.maths.IFunc(
            lambda a: (a, a ** 2.0), lambda cube: cf_units.Unit("1")
        )

        # should fail because data function returns a tuple
        with self.assertRaises(ValueError):
            my_ifunc(a)

        my_ifunc = iris.analysis.maths.IFunc(
            lambda a: math.sqrt(a), lambda cube: cf_units.Unit("1")
        )

        # should fail because data function does not work when its argument
        # is a numpy array
        with self.assertRaises(TypeError):
            my_ifunc(a)

    def test_type_error(self):
        with self.assertRaises(TypeError):
            iris.analysis.maths.add("not a cube", 123)


@tests.skip_data
class TestDivideAndMultiply(tests.IrisTest):
    def setUp(self):
        self.cube = iris.tests.stock.global_pp()
        self.cube.data = self.cube.data - 260

    def test_divide(self):
        a = self.cube

        c = a / a

        np.testing.assert_array_almost_equal(a.data / a.data, c.data)
        self.assertCML(c, ("analysis", "division.cml"), checksum=False)

        # Check that the division has had no effect on the original
        self.assertCML(a, ("analysis", "maths_original.cml"))

    def test_divide_by_scalar(self):
        a = self.cube

        c = a / 10

        np.testing.assert_array_almost_equal(a.data / 10, c.data)
        self.assertCML(c, ("analysis", "division_scalar.cml"), checksum=False)

        # Check that the division has had no effect on the original
        self.assertCML(a, ("analysis", "maths_original.cml"))

    def test_divide_by_coordinate(self):
        a = self.cube

        c = a / a.coord("latitude")
        self.assertCML(c, ("analysis", "division_by_latitude.cml"))

        # Check that the division has had no effect on the original
        self.assertCML(a, ("analysis", "maths_original.cml"))

    def test_divide_by_array(self):
        a = self.cube
        data_array = self.cube.copy().data

        # test division by exactly the same shape data
        c = a / data_array
        self.assertArrayEqual(c.data, np.array(1, dtype=np.float32))
        self.assertCML(
            c, ("analysis", "division_by_array.cml"), checksum=False
        )

        # test division by array of fewer dimensions
        c = a / data_array[0, :]
        self.assertArrayEqual(c.data[0, :], np.array(1, dtype=np.float32))

        # test division by array of more dimensions
        d_array = data_array.reshape(-1, data_array.shape[1], 1, 1)
        self.assertRaises(ValueError, iris.analysis.maths.divide, c, d_array)

        # Check that the division has had no effect on the original
        self.assertCML(a, ("analysis", "maths_original.cml"))

    def test_divide_by_coordinate_dim2(self):
        a = self.cube

        # Prevent divide-by-zero warning
        a.coord("longitude").points = a.coord("longitude").points + 0.5

        c = a / a.coord("longitude")
        self.assertCML(c, ("analysis", "division_by_longitude.cml"))

        # Reset to allow comparison with original
        a.coord("longitude").points = a.coord("longitude").points - 0.5

        # Check that the division has had no effect on the original
        self.assertCML(a, ("analysis", "maths_original.cml"))

    def test_divide_by_singular_coordinate(self):
        a = self.cube

        coord = DimCoord(points=2, long_name="foo", units="1")
        c = iris.analysis.maths.divide(a, coord)
        self.assertCML(c, ("analysis", "division_by_singular_coord.cml"))

        # Check that the division is equivalent to dividing the whole of the data by 2
        self.assertArrayEqual(c.data, a.data / 2.0)

    def test_divide_by_different_len_coord(self):
        a = self.cube

        coord = DimCoord(
            points=np.arange(10) * 2 + 5,
            standard_name="longitude",
            units="degrees",
        )

        self.assertRaises(ValueError, iris.analysis.maths.divide, a, coord)

    def test_divide_in_place(self):
        a = self.cube.copy()
        b = iris.analysis.maths.divide(a, 5, in_place=True)
        self.assertIs(a, b)

    def test_divide_not_in_place(self):
        a = self.cube.copy()
        b = iris.analysis.maths.divide(a, 5, in_place=False)
        self.assertIsNot(a, b)

    def test_multiply(self):
        a = self.cube

        c = a * a
        self.assertCML(c, ("analysis", "multiply.cml"))

        # Check that the multiplication has had no effect on the original
        self.assertCML(a, ("analysis", "maths_original.cml"))

    def test_multiplication_different_standard_name(self):
        a = self.cube.copy()
        b = self.cube.copy()
        b.rename("my cube data")
        c = a * b
        self.assertCML(
            c, ("analysis", "multiply_different_std_name.cml"), checksum=False
        )

    def test_multiplication_different_attributes(self):
        a = self.cube.copy()
        b = self.cube.copy()
        b.attributes["my attribute"] = "foobar"
        c = a * b
        self.assertIsNone(c.standard_name)
        expected = {
            "source": "Data from Met Office Unified Model",
            "my attribute": "foobar",
        }
        self.assertEqual(expected, c.attributes)

    def test_multiplication_in_place(self):
        a = self.cube.copy()
        b = iris.analysis.maths.multiply(a, 5, in_place=True)
        self.assertIs(a, b)

    def test_multiplication_not_in_place(self):
        a = self.cube.copy()
        b = iris.analysis.maths.multiply(a, 5, in_place=False)
        self.assertIsNot(a, b)


@tests.skip_data
class TestExponentiate(tests.IrisTest):
    def setUp(self):
        self.cube = iris.tests.stock.global_pp()
        # Increase dtype from float32 to float64 in order
        # to avoid dtype quantization errors during maths.
        self.cube.data = self.cube.data.astype(np.float64) - 260.0

    def test_exponentiate(self):
        a = self.cube
        e = pow(a, 4)
        self.assertCMLApproxData(e, ("analysis", "exponentiate.cml"))

    def test_square_root(self):
        # Make sure we have something which we can take the root of.
        a = self.cube
        a.data = abs(a.data)
        a.units **= 2

        e = a ** 0.5

        self.assertArrayAllClose(e.data, a.data ** 0.5)
        self.assertCML(e, ("analysis", "sqrt.cml"), checksum=False)
        self.assertRaises(ValueError, iris.analysis.maths.exponentiate, a, 0.3)

    def test_type_error(self):
        with self.assertRaises(TypeError):
            iris.analysis.maths.exponentiate("not a cube", 2)


class TestExponential(tests.IrisTest):
    def setUp(self):
        self.cube = iris.tests.stock.simple_1d()

    def test_exp(self):
        e = iris.analysis.maths.exp(self.cube)
        self.assertCMLApproxData(e, ("analysis", "exp.cml"))


class TestApplyUfunc(tests.IrisTest):
    def setUp(self):
        self.cube = iris.tests.stock.simple_2d()

    def test_apply_ufunc(self):
        a = self.cube
        a.units = cf_units.Unit("meters")

        b = iris.analysis.maths.apply_ufunc(
            np.square,
            a,
            new_name="more_thingness",
            new_unit=a.units ** 2,
            in_place=False,
        )

        ans = a.data ** 2

        self.assertArrayEqual(b.data, ans)
        self.assertEqual(b.name(), "more_thingness")
        self.assertEqual(b.units, cf_units.Unit("m^2"))

        def vec_mag(u, v):
            return math.sqrt(u ** 2 + v ** 2)

        c = a.copy() + 2

        vec_mag_ufunc = np.frompyfunc(vec_mag, 2, 1)
        b = iris.analysis.maths.apply_ufunc(vec_mag_ufunc, a, c)

        ans = a.data ** 2 + c.data ** 2
        b2 = b ** 2

        self.assertArrayAlmostEqual(b2.data, ans)


class TestIFunc(tests.IrisTest):
    def setUp(self):
        self.cube = iris.tests.stock.simple_2d()

    def test_ifunc(self):
        a = self.cube
        a.units = cf_units.Unit("meters")

        my_ifunc = iris.analysis.maths.IFunc(np.square, lambda x: x.units ** 2)
        b = my_ifunc(a, new_name="more_thingness", in_place=False)

        ans = a.data ** 2

        self.assertArrayEqual(b.data, ans)
        self.assertEqual(b.name(), "more_thingness")
        self.assertEqual(b.units, cf_units.Unit("m^2"))

        def vec_mag(u, v):
            return math.sqrt(u ** 2 + v ** 2)

        c = a.copy() + 2

        vec_mag_ufunc = np.frompyfunc(vec_mag, 2, 1)
        my_ifunc = iris.analysis.maths.IFunc(
            vec_mag_ufunc, lambda x, y: (x + y).units
        )
        b = my_ifunc(a, c)

        ans = (a.data ** 2 + c.data ** 2) ** 0.5

        self.assertArrayAlmostEqual(b.data, ans)

        def vec_mag_data_func(u_data, v_data):
            return np.sqrt(u_data ** 2 + v_data ** 2)

        vec_mag_ifunc = iris.analysis.maths.IFunc(
            vec_mag_data_func, lambda a, b: (a + b).units
        )
        b2 = vec_mag_ifunc(a, c)

        self.assertArrayAlmostEqual(b.data, b2.data)

        cs_ifunc = iris.analysis.maths.IFunc(np.cumsum, lambda a: a.units)

        b = cs_ifunc(a, axis=1)
        ans = a.data.copy()
        ans = np.cumsum(ans, axis=1)

        self.assertArrayAlmostEqual(b.data, ans)


@tests.skip_data
class TestLog(tests.IrisTest):
    def setUp(self):
        self.cube = iris.tests.stock.global_pp()

    def test_log(self):
        e = iris.analysis.maths.log(self.cube)
        self.assertCMLApproxData(e, ("analysis", "log.cml"))

    def test_log2(self):
        e = iris.analysis.maths.log2(self.cube)
        self.assertCMLApproxData(e, ("analysis", "log2.cml"))

    def test_log10(self):
        e = iris.analysis.maths.log10(self.cube)
        self.assertCMLApproxData(e, ("analysis", "log10.cml"), rtol=1e-6)


class TestMathOperations(tests.IrisTest):
    def setUp(self):
        self.data_1f = np.array([[9, 9, 9], [8, 8, 8]], dtype=np.float64)
        self.data_2f = np.array([[3, 3, 3], [2, 2, 2]], dtype=np.float64)
        self.data_1i = np.array([[9, 9, 9], [8, 8, 8]], dtype=np.int64)
        self.data_2i = np.array([[3, 3, 3], [2, 2, 2]], dtype=np.int64)
        self.data_1u = np.array([[9, 9, 9], [8, 8, 8]], dtype=np.uint64)
        self.data_2u = np.array([[3, 3, 3], [2, 2, 2]], dtype=np.uint64)

        self.cube_1f = Cube(self.data_1f)
        self.cube_2f = Cube(self.data_2f)
        self.cube_1i = Cube(self.data_1i)
        self.cube_2i = Cube(self.data_2i)
        self.cube_1u = Cube(self.data_1u)
        self.cube_2u = Cube(self.data_2u)

        self.ops = (operator.add, operator.sub, operator.mul, operator.truediv)
        self.iops = (
            operator.iadd,
            operator.isub,
            operator.imul,
            operator.itruediv,
        )

    def common_neg(self, cube, data):
        result1 = -cube
        result2 = -data
        self.assertIsInstance(result1, Cube)
        self.assertIsNot(result1, cube)
        self.assertArrayAlmostEqual(result1.data, result2)

    def test_neg_f(self):
        self.common_neg(self.cube_1f, self.data_1f)

    def test_neg_i(self):
        self.common_neg(self.cube_1i, self.data_1i)

    def test_neg_u(self):
        self.common_neg(self.cube_1u, self.data_1u)

    def test_operator(self):
        for test_op in self.ops:
            result1 = test_op(self.cube_1f, self.cube_2f)
            result2 = test_op(self.data_1f, self.data_2f)
            self.assertArrayAlmostEqual(result1.data, result2)
            result3 = test_op(self.cube_1i, self.cube_2i)
            result4 = test_op(self.data_1i, self.data_2i)
            self.assertArrayAlmostEqual(result3.data, result4)
            result3 = test_op(self.cube_1u, self.cube_2u)
            result4 = test_op(self.data_1u, self.data_2u)
            self.assertArrayAlmostEqual(result3.data, result4)

    def test_operator__mixed_floats_ints(self):
        for test_op in self.ops:
            result1 = test_op(self.cube_1f, self.cube_2i)
            result2 = test_op(self.data_1f, self.data_2i)
            self.assertArrayAlmostEqual(result1.data, result2)
            result3 = test_op(self.cube_1i, self.cube_2f)
            result4 = test_op(self.data_1i, self.data_2f)
            self.assertArrayAlmostEqual(result3.data, result4)

    def test_operator__inplace(self):
        for test_op in self.iops:
            test_op(self.cube_1f, self.cube_2f)
            test_op(self.data_1f, self.data_2f)
            self.assertArrayAlmostEqual(self.cube_1f.data, self.data_1f)
            if test_op != operator.itruediv:
                # You can't safely cast floats from true-dividing back to ints
                # to satisfy the in-place requirement.
                test_op(self.cube_1i, self.cube_2i)
                expected = test_op(self.data_1i, self.data_2i)
                self.assertArrayAlmostEqual(self.cube_1i.data, expected)
                test_op(self.cube_1u, self.cube_2u)
                expected = test_op(self.data_1u, self.data_2u)
                self.assertArrayAlmostEqual(self.cube_1u.data, expected)

    def test_operator__inplace__mixed_floats_ints_uints(self):
        # Can only test float-type numerators as int-type numerators
        # would require casting ints to floats to satisfy the in-place
        # requirement, which cannot be done 'safely'.
        for test_op in self.iops:
            result1 = test_op(self.cube_1f, self.cube_2i)
            result2 = test_op(self.data_1f, self.data_2i)
            self.assertArrayAlmostEqual(result1.data, result2)
            result1 = test_op(self.cube_1f, self.cube_2u)
            result2 = test_op(self.data_1f, self.data_2u)
            self.assertArrayAlmostEqual(result1.data, result2)

    def test_operator__inplace__mixed_ints_uints_floats(self):
        # Math operations on int-type numerators float-type denominators
        # would require casting floats to ints to satisfy the in-place
        # requirement.
        for test_op in self.iops:
            test_emsg = "Cannot perform inplace"
            with self.assertRaisesRegex(ArithmeticError, test_emsg):
                test_op(self.cube_1i, self.cube_2f)
            with self.assertRaisesRegex(ArithmeticError, test_emsg):
                test_op(self.cube_1u, self.cube_2f)

    def test_operator__inplace__scalar_int(self):
        # Math operations on int-type numerators float-type denominators
        # would require casting floats to ints to satisfy the in-place
        # requirement.
        scalar = 2
        for test_op in self.iops:
            if test_op != operator.itruediv:
                # You can't safely cast floats from true-dividing back to ints
                # to satisfy the in-place requirement.
                test_op(self.cube_1i, scalar)
                test_op(self.data_1i, scalar)
                self.assertArrayAlmostEqual(self.cube_1i.data, self.data_1i)
                test_op(self.cube_1u, scalar)
                test_op(self.data_1u, scalar)
                self.assertArrayAlmostEqual(self.cube_1u.data, self.data_1u)

    def test_operator__inplace__scalar_float(self):
        # Math operations on int-type numerators float-type denominators
        # would require casting floats to ints to satisfy the in-place
        # requirement.
        scalar = 2.5
        for test_op in self.iops:
            test_emsg = "Cannot perform inplace"
            with self.assertRaisesRegex(ArithmeticError, test_emsg):
                test_op(self.cube_1i, scalar)
            with self.assertRaisesRegex(ArithmeticError, test_emsg):
                test_op(self.cube_1u, scalar)

    def test_operator__scalar_float(self):
        scalar = 2.5
        for test_op in self.ops:
            result1 = test_op(self.cube_1f, scalar)
            result2 = test_op(self.data_1f, scalar)
            self.assertArrayAlmostEqual(result1.data, result2)
            result3 = test_op(self.cube_1i, scalar)
            result4 = test_op(self.data_1i, scalar)
            self.assertArrayAlmostEqual(result3.data, result4)

    def test_operator_scalar__int(self):
        scalar = 2
        for test_op in self.ops:
            result1 = test_op(self.cube_1f, scalar)
            result2 = test_op(self.data_1f, scalar)
            self.assertArrayAlmostEqual(result1.data, result2)
            result3 = test_op(self.cube_1i, scalar)
            result4 = test_op(self.data_1i, scalar)
            self.assertArrayAlmostEqual(result3.data, result4)

    def test_operator_array(self):
        for test_op in self.ops:
            result1 = test_op(self.cube_1f, self.data_2f)
            result2 = test_op(self.data_1f, self.data_2f)
            self.assertArrayAlmostEqual(result1.data, result2)
            result3 = test_op(self.cube_1i, self.data_2i)
            result4 = test_op(self.data_1i, self.data_2i)
            self.assertArrayAlmostEqual(result3.data, result4)

    def test_cube_itruediv__int(self):
        with self.assertRaisesRegex(
            ArithmeticError, "Cannot perform inplace division"
        ):
            operator.itruediv(self.cube_1i, self.cube_2i)

    def test_cube_itruediv__uint(self):
        with self.assertRaisesRegex(
            ArithmeticError, "Cannot perform inplace division"
        ):
            operator.itruediv(self.cube_1u, self.cube_2u)

    def test_int_cube_itruediv__scalar(self):
        with self.assertRaisesRegex(
            ArithmeticError, "Cannot perform inplace division"
        ):
            operator.itruediv(self.cube_1i, 5)

    def test_uint_cube_itruediv__scalar(self):
        with self.assertRaisesRegex(
            ArithmeticError, "Cannot perform inplace division"
        ):
            operator.itruediv(self.cube_1u, 5)


class TestMaskedArrays(tests.IrisTest):
    def setUp(self):
        self.data = ma.MaskedArray(
            [[9, 9, 9], [8, 8, 8]],
            mask=[[0, 1, 0], [0, 0, 1]],
            dtype=np.float64,
        )
        self.cube = Cube(self.data)

    def test_incompatible_dimensions(self):
        data3 = ma.MaskedArray(
            [[3, 3, 3, 4], [2, 2, 2, 2]], mask=[[0, 1, 0, 0], [0, 1, 1, 1]]
        )
        with self.assertRaises(ValueError):
            # Incompatible dimensions.
            self.cube + data3

    def test_increase_cube_dimensionality(self):
        with self.assertRaises(ValueError):
            # This would increase the dimensionality of the cube
            # due to auto-broadcasting.
            cube_x = Cube(ma.MaskedArray([[9]], mask=[[0]]))
            cube_x + ma.MaskedArray([[3, 3, 3, 3]], mask=[[0, 1, 0, 1]])


class TestCoordMathOperations(tests.IrisTest):
    def setUp(self):
        self.value_cube = 100
        self.value = 10
        self.cube = Cube([self.value_cube])
        self.dim = DimCoord([self.value])
        self.aux = AuxCoord([self.value])
        self.i = int(self.value)
        self.f = float(self.value)
        self.np_i = np.int32(self.value)
        self.np_f = np.float32(self.value)
        self.numbers = (self.i, self.f, self.np_i, self.np_f)
        self.ops = (operator.add, operator.sub, operator.mul, operator.truediv)
        self.symbols = ("+", "-", "*", "/")

    def test_coord_op_coord__fail(self):
        for op, symbol in zip(self.ops, self.symbols):
            emsg = f"AuxCoord \{symbol} DimCoord"  # noqa: W605
            with self.assertRaisesRegex(NotYetImplementedError, emsg):
                _ = op(self.aux, self.dim)

    def test_coord_op_number(self):
        for op in self.ops:
            for number in self.numbers:
                actual = op(self.aux, number)
                expected = op(self.value, number)
                self.assertIsInstance(actual, AuxCoord)
                self.assertEqual(expected, actual.points)

    def test_number_op_coord(self):
        for op in self.ops:
            for number in self.numbers:
                actual = op(number, self.dim)
                expected = op(number, self.value)
                self.assertIsInstance(actual, DimCoord)
                self.assertEqual(expected, actual.points)

    def test_coord_op_cube(self):
        for op in self.ops:
            actual = op(self.aux, self.cube)
            expected = op(self.value, self.value_cube)
            self.assertIsInstance(actual, Cube)
            self.assertEqual(expected, actual.data)

    def test_cube_op_coord(self):
        for op in self.ops:
            actual = op(self.cube, self.dim)
            expected = op(self.value_cube, self.value)
            self.assertIsInstance(actual, Cube)
            self.assertEqual(actual.data, expected)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for :class:`iris.fileformat.ff.Grid`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

from iris.fileformats._ff import Grid


class Test___init__(tests.IrisTest):
    def test_attributes(self):
        # Ensure the constructor initialises all the grid's attributes
        # correctly, including unpacking values from the REAL constants.
        reals = (
            mock.sentinel.ew,
            mock.sentinel.ns,
            mock.sentinel.first_lat,
            mock.sentinel.first_lon,
            mock.sentinel.pole_lat,
            mock.sentinel.pole_lon,
        )
        grid = Grid(
            mock.sentinel.column,
            mock.sentinel.row,
            reals,
            mock.sentinel.horiz_grid_type,
        )
        self.assertIs(grid.column_dependent_constants, mock.sentinel.column)
        self.assertIs(grid.row_dependent_constants, mock.sentinel.row)
        self.assertIs(grid.ew_spacing, mock.sentinel.ew)
        self.assertIs(grid.ns_spacing, mock.sentinel.ns)
        self.assertIs(grid.first_lat, mock.sentinel.first_lat)
        self.assertIs(grid.first_lon, mock.sentinel.first_lon)
        self.assertIs(grid.pole_lat, mock.sentinel.pole_lat)
        self.assertIs(grid.pole_lon, mock.sentinel.pole_lon)
        self.assertIs(grid.horiz_grid_type, mock.sentinel.horiz_grid_type)


class Test_vectors(tests.IrisTest):
    def setUp(self):
        self.xp = mock.sentinel.xp
        self.xu = mock.sentinel.xu
        self.yp = mock.sentinel.yp
        self.yv = mock.sentinel.yv

    def _test_subgrid_vectors(self, subgrid, expected):
        grid = Grid(None, None, (None,) * 6, None)
        grid._x_vectors = mock.Mock(return_value=(self.xp, self.xu))
        grid._y_vectors = mock.Mock(return_value=(self.yp, self.yv))
        result = grid.vectors(subgrid)
        self.assertEqual(result, expected)

    def test_1(self):
        # Data on atmospheric theta points.
        self._test_subgrid_vectors(1, (self.xp, self.yp))

    def test_2(self):
        # Data on atmospheric theta points, values over land only.
        self._test_subgrid_vectors(2, (self.xp, self.yp))

    def test_3(self):
        # Data on atmospheric theta points, values over sea only.
        self._test_subgrid_vectors(3, (self.xp, self.yp))

    def test_4(self):
        # Data on atmospheric zonal theta points.
        self._test_subgrid_vectors(4, (self.xp, self.yp))

    def test_5(self):
        # Data on atmospheric meridional theta points.
        self._test_subgrid_vectors(5, (self.xp, self.yp))

    def test_11(self):
        # Data on atmospheric uv points.
        self._test_subgrid_vectors(11, (self.xu, self.yv))

    def test_18(self):
        # Data on atmospheric u points on the 'c' grid.
        self._test_subgrid_vectors(18, (self.xu, self.yp))

    def test_19(self):
        # Data on atmospheric v points on the 'c' grid.
        self._test_subgrid_vectors(19, (self.xp, self.yv))

    def test_26(self):
        # Lateral boundary data at atmospheric theta points.
        self._test_subgrid_vectors(26, (self.xp, self.yp))

    def test_27(self):
        # Lateral boundary data at atmospheric u points.
        self._test_subgrid_vectors(27, (self.xu, self.yp))

    def test_28(self):
        # Lateral boundary data at atmospheric v points.
        self._test_subgrid_vectors(28, (self.xp, self.yv))

    def test_29(self):
        # Orography field for atmospheric LBCs.
        self._test_subgrid_vectors(29, (self.xp, self.yp))


if __name__ == "__main__":
    tests.main()



class ExampleClass:
    """
    Class Summary

    """

    def __init__(self, arg1, arg2):
        """
        Purpose section description.

        Description section text.

        Args:

        * arg1 (int):
            First argument description.
        * arg2 (float):
            Second argument description.

        Returns:
            Boolean.

        """
        self.a = arg1
        "Attribute arg1 docstring."
        self.b = arg2
        "Attribute arg2 docstring."

    @property
    def square(self):
        """
        *(read-only)* Purpose section description.

        Returns:
            int.

        """
        return self.a * self.a



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Provides NIMROD file format capabilities."""

import glob
import numpy as np
import os
import struct
import sys

import iris
from iris.exceptions import TranslationError
import iris.fileformats.nimrod_load_rules


# general header (int16) elements 1-31 (Fortran bytes 1-62)
general_header_int16s = (
    "vt_year",
    "vt_month",
    "vt_day",
    "vt_hour",
    "vt_minute",
    "vt_second",
    "dt_year",
    "dt_month",
    "dt_day",
    "dt_hour",
    "dt_minute",
    "datum_type",
    "datum_len",
    "experiment_num",
    "horizontal_grid_type",
    "num_rows",
    "num_cols",
    "nimrod_version",
    "field_code",
    "vertical_coord_type",
    "reference_vertical_coord_type",
    "data_specific_float32_len",
    "data_specific_int16_len",
    "origin_corner",
    "int_mdi",
    "period_minutes",
    "num_model_levels",
    "proj_biaxial_ellipsoid",
    "ensemble_member",
    "model_origin_id",
    "averagingtype",
)


# general header (float32) elements 32-59 (Fortran bytes 63-174)
general_header_float32s = (
    "vertical_coord",
    "reference_vertical_coord",
    "y_origin",
    "row_step",
    "x_origin",
    "column_step",
    "float32_mdi",
    "MKS_data_scaling",
    "data_offset",
    "x_offset",
    "y_offset",
    "true_origin_latitude",
    "true_origin_longitude",
    "true_origin_easting",
    "true_origin_northing",
    "tm_meridian_scaling",
    "threshold_value_alt",
    "threshold_value",
)


# data specific header (float32) elements 60-104 (Fortran bytes 175-354)
data_header_float32s = (
    "tl_y",
    "tl_x",
    "tr_y",
    "tr_x",
    "br_y",
    "br_x",
    "bl_y",
    "bl_x",
    "sat_calib",
    "sat_space_count",
    "ducting_index",
    "elevation_angle",
    "neighbourhood_radius",
    "threshold_vicinity_radius",
    "recursive_filter_alpha",
    "threshold_fuzziness",
    "threshold_duration_fuzziness",
)


# data specific header (char) elements 105-107 (bytes 355-410)
# units, source and title


# data specific header (int16) elements 108-159 (Fortran bytes 411-512)
data_header_int16s = (
    "threshold_type",
    "probability_method",
    "recursive_filter_iterations",
    "member_count",
    "probability_period_of_event",
    "data_header_int16_05",
    "soil_type",
    "radiation_code",
    "data_header_int16_08",
    "data_header_int16_09",
    "data_header_int16_10",
    "data_header_int16_11",
    "data_header_int16_12",
    "data_header_int16_13",
    "data_header_int16_14",
    "data_header_int16_15",
    "data_header_int16_16",
    "data_header_int16_17",
    "data_header_int16_18",
    "data_header_int16_19",
    "data_header_int16_20",
    "data_header_int16_21",
    "data_header_int16_22",
    "data_header_int16_23",
    "data_header_int16_24",
    "data_header_int16_25",
    "data_header_int16_26",
    "data_header_int16_27",
    "data_header_int16_28",
    "data_header_int16_29",
    "data_header_int16_30",
    "data_header_int16_31",
    "data_header_int16_32",
    "data_header_int16_33",
    "data_header_int16_34",
    "data_header_int16_35",
    "data_header_int16_36",
    "data_header_int16_37",
    "data_header_int16_38",
    "data_header_int16_39",
    "data_header_int16_40",
    "data_header_int16_41",
    "data_header_int16_42",
    "data_header_int16_43",
    "data_header_int16_44",
    "data_header_int16_45",
    "data_header_int16_46",
    "data_header_int16_47",
    "data_header_int16_48",
    "data_header_int16_49",
    "period_seconds",
)


def _read_chars(infile, num):
    """Read characters from the (big-endian) file."""
    instr = infile.read(num)
    result = struct.unpack(">%ds" % num, instr)[0]
    result = result.decode()
    return result


class NimrodField:
    """
    A data field from a NIMROD file.

    Capable of converting itself into a :class:`~iris.cube.Cube`

    References:
        Met Office (2003): Met Office Rain Radar Data from the NIMROD System.
        NCAS British Atmospheric Data Centre, date of citation.
        http://catalogue.ceda.ac.uk/uuid/82adec1f896af6169112d09cc1174499

    """

    def __init__(self, from_file=None):
        """
        Create a NimrodField object and optionally read from an open file.

        Example::

            with open("nimrod_file", "rb") as infile:
                field = NimrodField(infile)

        """
        if from_file is not None:
            self.read(from_file)

    def read(self, infile):
        """Read the next field from the given file object."""
        self._read_header(infile)
        self._read_data(infile)

    def _read_header_subset(self, infile, names, dtype):
        # Read contiguous header items of the same data type.
        values = np.fromfile(infile, dtype=dtype, count=len(names))
        if sys.byteorder == "little":
            values.byteswap(True)
        for i, name in enumerate(names):
            setattr(self, name, values[i])

    def _read_header(self, infile):
        """Load the 512 byte header (surrounded by 4-byte length)."""

        leading_length = struct.unpack(">L", infile.read(4))[0]
        if leading_length != 512:
            raise TranslationError("Expected header leading_length of 512")

        # general header (int16) elements 1-31 (bytes 1-62)
        self._read_header_subset(infile, general_header_int16s, np.int16)

        # general header (float32) elements 32-59 (bytes 63-174)
        self._read_header_subset(infile, general_header_float32s, np.float32)
        # skip unnamed floats
        infile.seek(4 * (28 - len(general_header_float32s)), os.SEEK_CUR)

        # data specific header (float32) elements 60-104 (bytes 175-354)
        self._read_header_subset(infile, data_header_float32s, np.float32)
        # skip unnamed floats
        infile.seek(4 * (45 - len(data_header_float32s)), os.SEEK_CUR)

        # data specific header (char) elements 105-107 (bytes 355-410)
        self.units = _read_chars(infile, 8)
        self.source = _read_chars(infile, 24)
        self.title = _read_chars(infile, 24)

        # data specific header (int16) elements 108- (bytes 411-512)
        self._read_header_subset(infile, data_header_int16s, np.int16)
        # skip unnamed int16s
        infile.seek(2 * (51 - len(data_header_int16s)), os.SEEK_CUR)

        trailing_length = struct.unpack(">L", infile.read(4))[0]
        if trailing_length != leading_length:
            raise TranslationError(
                "Expected header trailing_length of {}, "
                "got {}.".format(leading_length, trailing_length)
            )

    def _read_data(self, infile):
        """
        Read the data array: int8, int16, int32 or float32

        (surrounded by 4-byte length, at start and end)

        """
        # what are we expecting?
        num_data = int(self.num_rows) * int(self.num_cols)
        num_data_bytes = int(num_data) * int(self.datum_len)

        # format string for unpacking the file.read()
        # 0:real
        if self.datum_type == 0:
            numpy_dtype = np.float32
        # 1:int
        elif self.datum_type == 1:
            if self.datum_len == 1:
                numpy_dtype = np.int8
            elif self.datum_len == 2:
                numpy_dtype = np.int16
            elif self.datum_len == 4:
                numpy_dtype = np.int32
            else:
                raise TranslationError(
                    "Undefined datum length " "%d" % self.datum_type
                )
        # 2:byte
        elif self.datum_type == 2:
            numpy_dtype = np.byte
        else:
            raise TranslationError("Undefined data type")
        leading_length = struct.unpack(">L", infile.read(4))[0]
        if leading_length != num_data_bytes:
            raise TranslationError(
                "Expected data leading_length of %d" % num_data_bytes
            )

        self.data = np.fromfile(infile, dtype=numpy_dtype, count=num_data)

        if sys.byteorder == "little":
            self.data.byteswap(True)

        trailing_length = struct.unpack(">L", infile.read(4))[0]
        if trailing_length != leading_length:
            raise TranslationError(
                "Expected data trailing_length of %d" % num_data_bytes
            )

        # Form the correct shape.
        self.data = self.data.reshape(self.num_rows, self.num_cols)


def load_cubes(filenames, callback=None):
    """
    Loads cubes from a list of NIMROD filenames.

    Args:

    * filenames - list of NIMROD filenames to load

    Kwargs:

    * callback - a function which can be passed on to
                 :func:`iris.io.run_callback`

    .. note::

        The resultant cubes may not be in the same order as in the files.

    """
    if isinstance(filenames, str):
        filenames = [filenames]

    for filename in filenames:
        for path in glob.glob(filename):
            with open(path, "rb") as infile:
                while True:
                    try:
                        field = NimrodField(infile)
                    except struct.error:
                        # End of file. Move on to the next file.
                        break

                    cube = iris.fileformats.nimrod_load_rules.run(field)

                    # Were we given a callback?
                    if callback is not None:
                        cube = iris.io.run_callback(
                            callback, cube, field, filename
                        )
                    if cube is None:
                        continue

                    yield cube



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for :func:`iris.fileformats.name_loaders._build_cell_methods`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

import iris.coords
from iris.fileformats.name_loaders import _build_cell_methods


class Tests(tests.IrisTest):
    def test_nameII_average(self):
        av_or_int = ["something average ob bla"]
        coord_name = "foo"
        res = _build_cell_methods(av_or_int, coord_name)
        self.assertEqual(res, [iris.coords.CellMethod("mean", "foo")])

    def test_nameIII_averaged(self):
        av_or_int = ["something averaged ob bla"]
        coord_name = "bar"
        res = _build_cell_methods(av_or_int, coord_name)
        self.assertEqual(res, [iris.coords.CellMethod("mean", "bar")])

    def test_nameII_integral(self):
        av_or_int = ["something integral ob bla"]
        coord_name = "ensemble"
        res = _build_cell_methods(av_or_int, coord_name)
        self.assertEqual(res, [iris.coords.CellMethod("sum", "ensemble")])

    def test_nameIII_integrated(self):
        av_or_int = ["something integrated ob bla"]
        coord_name = "time"
        res = _build_cell_methods(av_or_int, coord_name)
        self.assertEqual(res, [iris.coords.CellMethod("sum", "time")])

    def test_no_averaging(self):
        av_or_int = [
            "No foo averaging",
            "No bar averaging",
            "no",
            "",
            "no averaging",
            "no anything at all averaging",
        ]
        coord_name = "time"
        res = _build_cell_methods(av_or_int, coord_name)
        self.assertEqual(res, [None] * len(av_or_int))

    def test_nameII_mixed(self):
        av_or_int = [
            "something integral ob bla",
            "no averaging",
            "other average",
        ]
        coord_name = "ensemble"
        res = _build_cell_methods(av_or_int, coord_name)
        self.assertEqual(
            res,
            [
                iris.coords.CellMethod("sum", "ensemble"),
                None,
                iris.coords.CellMethod("mean", "ensemble"),
            ],
        )

    def test_nameIII_mixed(self):
        av_or_int = [
            "something integrated ob bla",
            "no averaging",
            "other averaged",
        ]
        coord_name = "ensemble"
        res = _build_cell_methods(av_or_int, coord_name)
        self.assertEqual(
            res,
            [
                iris.coords.CellMethod("sum", "ensemble"),
                None,
                iris.coords.CellMethod("mean", "ensemble"),
            ],
        )

    def test_unrecognised(self):
        unrecognised_heading = "bla else"
        av_or_int = [
            "something average",
            unrecognised_heading,
            "something integral",
        ]
        coord_name = "foo"
        with mock.patch("warnings.warn") as warn:
            _ = _build_cell_methods(av_or_int, coord_name)
        expected_msg = (
            "Unknown {} statistic: {!r}. Unable to "
            "create cell method.".format(coord_name, unrecognised_heading)
        )
        warn.assert_called_with(expected_msg)

    def test_unrecognised_similar_to_no_averaging(self):
        unrecognised_headings = [
            "not averaging",
            "this is not a valid no",
            "nope",
            "no daveraging",
            "no averagingg",
            "no something",
            "noaveraging",
        ]
        for unrecognised_heading in unrecognised_headings:
            av_or_int = [
                "something average",
                unrecognised_heading,
                "something integral",
            ]
            coord_name = "foo"
            with mock.patch("warnings.warn") as warn:
                _ = _build_cell_methods(av_or_int, coord_name)
            expected_msg = (
                "Unknown {} statistic: {!r}. Unable to "
                "create cell method.".format(coord_name, unrecognised_heading)
            )
            warn.assert_called_with(expected_msg)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.plot._check_bounds_contiguity_and_mask`
function."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

import numpy as np
import numpy.ma as ma

from iris.coords import DimCoord
from iris.tests.stock import (
    sample_2d_latlons,
    make_bounds_discontiguous_at_point,
)

from iris.plot import _check_bounds_contiguity_and_mask


@tests.skip_plot
class Test_check_bounds_contiguity_and_mask(tests.IrisTest):
    def test_1d_not_checked(self):
        # Test a 1D coordinate, which is not checked as atol is not set.
        coord = DimCoord([1, 3, 5], bounds=[[0, 2], [2, 4], [5, 6]])
        data = np.array([278, 300, 282])
        # Make sure contiguity checking doesn't throw an error
        _check_bounds_contiguity_and_mask(coord, data)

    def test_1d_contiguous(self):
        # Test that a 1D coordinate which is contiguous does not fail.
        coord = DimCoord([1, 3, 5], bounds=[[0, 2], [2, 4], [4, 6]])
        data = np.array([278, 300, 282])
        _check_bounds_contiguity_and_mask(coord, data, atol=1e-3)

    def test_1d_discontigous_masked(self):
        # Test a 1D coordinate which is discontiguous but masked at
        # discontiguities.
        coord = DimCoord([1, 3, 5], bounds=[[0, 2], [2, 4], [5, 6]])
        data = ma.array(np.array([278, 300, 282]), mask=[0, 1, 0])
        _check_bounds_contiguity_and_mask(coord, data, atol=1e-3)

    def test_1d_discontigous_unmasked(self):
        # Test a 1D coordinate which is discontiguous and unmasked at
        # discontiguities.
        coord = DimCoord([1, 3, 5], bounds=[[0, 2], [2, 4], [5, 6]])
        data = ma.array(np.array([278, 300, 282]), mask=[1, 0, 0])
        msg = (
            "coordinate are not contiguous and data is not masked where "
            "the discontiguity occurs"
        )
        with self.assertRaisesRegex(ValueError, msg):
            _check_bounds_contiguity_and_mask(coord, data, atol=1e-3)

    def test_2d_contiguous(self):
        # Test that a 2D coordinate which is contiguous does not throw
        # an error.
        cube = sample_2d_latlons()
        _check_bounds_contiguity_and_mask(cube.coord("longitude"), cube.data)

    def test_2d_contiguous_atol(self):
        # Check the atol is passed correctly.
        cube = sample_2d_latlons()
        with mock.patch(
            "iris.coords.Coord._discontiguity_in_bounds"
        ) as discontiguity_check:
            # Discontiguity returns two objects that are unpacked in
            # `_check_bounds_contiguity_and_mask`.
            discontiguity_check.return_value = [True, None]
            _check_bounds_contiguity_and_mask(
                cube.coord("longitude"), cube.data, atol=1e-3
            )
        discontiguity_check.assert_called_with(atol=1e-3)

    def test_2d_discontigous_masked(self):
        # Test that a 2D coordinate which is discontiguous but masked at
        # discontiguities doesn't error.
        cube = sample_2d_latlons()
        make_bounds_discontiguous_at_point(cube, 3, 4)
        _check_bounds_contiguity_and_mask(cube.coord("longitude"), cube.data)

    def test_2d_discontigous_unmasked(self):
        # Test a 2D coordinate which is discontiguous and unmasked at
        # discontiguities.
        cube = sample_2d_latlons()
        make_bounds_discontiguous_at_point(cube, 3, 4)
        msg = "coordinate are not contiguous"
        cube.data[3, 4] = ma.nomask
        with self.assertRaisesRegex(ValueError, msg):
            _check_bounds_contiguity_and_mask(
                cube.coord("longitude"), cube.data
            )


if __name__ == "__main__":
    tests.main()
